"""
#exonware/xwlazy/src/exonware/xwlazy/discovery/discovery.py

Package discovery implementation.

Company: eXonware.com
Author: Eng. Muhammad AlShehri
Email: connect@exonware.com

Generation Date: 10-Oct-2025

This module provides LazyDiscovery class that discovers dependencies from
project configuration sources with caching support.
"""

import json
import re
import subprocess
import sys
import threading
from pathlib import Path
from typing import Dict, List, Optional

from ..base import APackageHelper
from ...defs import DependencyInfo
from ...common.logger import get_logger, log_event as _log

logger = get_logger("xwlazy.discovery")

class LazyDiscovery(APackageHelper):
    """
    Discovers dependencies from project configuration sources.
    Implements caching with file modification time checks.
    """
    
    # System/built-in modules that should NEVER be auto-installed
    SYSTEM_MODULES_BLACKLIST = {
        'pwd', 'grp', 'spwd', 'crypt', 'nis', 'syslog', 'termios', 'tty', 'pty',
        'fcntl', 'resource', 'msvcrt', 'winreg', 'winsound', '_winapi',
        'rpython', 'rply', 'rnc2rng', '_dbm',
        'sys', 'os', 'io', 'time', 'datetime', 'json', 'csv', 'math',
        'random', 're', 'collections', 'itertools', 'functools', 'operator',
        'pathlib', 'shutil', 'glob', 'tempfile', 'pickle', 'copy', 'types',
        'typing', 'abc', 'enum', 'dataclasses', 'contextlib', 'warnings',
        'logging', 'threading', 'multiprocessing', 'subprocess', 'queue',
        'socket', 'select', 'signal', 'asyncio', 'concurrent', 'email',
        'http', 'urllib', 'xml', 'html', 'sqlite3', 'base64', 'hashlib',
        'hmac', 'secrets', 'ssl', 'binascii', 'struct', 'array', 'weakref',
        'gc', 'inspect', 'traceback', 'atexit', 'codecs', 'locale', 'gettext',
        'argparse', 'optparse', 'configparser', 'fileinput', 'stat', 'platform',
        'unittest', 'doctest', 'pdb', 'profile', 'cProfile', 'timeit', 'trace',
        # Internal / optional modules that must never trigger auto-install
        'compression', 'socks', 'wimlib',
    }
    
    # Common import name to package name mappings
    COMMON_MAPPINGS = {
        'cv2': 'opencv-python',
        'PIL': 'Pillow',
        'Pillow': 'Pillow',
        'yaml': 'PyYAML',
        'sklearn': 'scikit-learn',
        'bs4': 'beautifulsoup4',
        'dateutil': 'python-dateutil',
        'requests_oauthlib': 'requests-oauthlib',
        'google': 'google-api-python-client',
        'jwt': 'PyJWT',
        'crypto': 'pycrypto',
        'Crypto': 'pycrypto',
        'MySQLdb': 'mysqlclient',
        'psycopg2': 'psycopg2-binary',
        'bson': 'pymongo',
        'lxml': 'lxml',
        'numpy': 'numpy',
        'pandas': 'pandas',
        'matplotlib': 'matplotlib',
        'seaborn': 'seaborn',
        'plotly': 'plotly',
        'django': 'Django',
        'flask': 'Flask',
        'fastapi': 'fastapi',
        'uvicorn': 'uvicorn',
        'pytest': 'pytest',
        'black': 'black',
        'isort': 'isort',
        'mypy': 'mypy',
        'psutil': 'psutil',
        'colorama': 'colorama',
        'pytz': 'pytz',
        'aiofiles': 'aiofiles',
        'watchdog': 'watchdog',
        'wand': 'Wand',
        'exifread': 'ExifRead',
        'piexif': 'piexif',
        'rawpy': 'rawpy',
        'imageio': 'imageio',
        'scipy': 'scipy',
        'scikit-image': 'scikit-image',
        'opencv-python': 'opencv-python',
        'opencv-contrib-python': 'opencv-contrib-python',
        'opentelemetry': 'opentelemetry-api',
        'opentelemetry.trace': 'opentelemetry-api',
        'opentelemetry.sdk': 'opentelemetry-sdk',
    }
    
    def _discover_from_sources(self) -> None:
        """Discover dependencies from all sources."""
        self._discover_from_pyproject_toml()
        self._discover_from_requirements_txt()
        self._discover_from_setup_py()
        self._discover_from_custom_config()
        self._add_common_mappings()  # Add well-known mappings (bson->pymongo, cv2->opencv-python, etc.)
    
    def _is_cache_valid(self) -> bool:
        """Check if cached dependencies are still valid."""
        if not self._cache_valid or not self._cached_dependencies:
            return False
        
        config_files = [
            self.project_root / 'pyproject.toml',
            self.project_root / 'requirements.txt',
            self.project_root / 'setup.py',
        ]
        
        for config_file in config_files:
            if config_file.exists():
                try:
                    current_mtime = config_file.stat().st_mtime
                    cached_mtime = self._file_mtimes.get(str(config_file), 0)
                    if current_mtime > cached_mtime:
                        return False
                except Exception:
                    return False
        
        return True
    
    def _update_file_mtimes(self) -> None:
        """Update file modification times for cache validation."""
        config_files = [
            self.project_root / 'pyproject.toml',
            self.project_root / 'requirements.txt',
            self.project_root / 'setup.py',
        ]
        for config_file in config_files:
            if config_file.exists():
                try:
                    self._file_mtimes[str(config_file)] = config_file.stat().st_mtime
                except Exception:
                    pass
    
    def _discover_from_pyproject_toml(self) -> None:
        """Discover dependencies from pyproject.toml."""
        pyproject_path = self.project_root / 'pyproject.toml'
        if not pyproject_path.exists():
            return
        
        try:
            try:
                import tomllib  # Python 3.11+
                toml_parser = tomllib  # type: ignore[assignment]
            except ImportError:
                try:
                    import tomli as tomllib  # type: ignore[assignment]
                    toml_parser = tomllib
                except ImportError:
                    _log(
                        "discovery",
                        "TOML parser not available; attempting to lazy-install 'tomli'...",
                    )
                    try:
                        subprocess.run(
                            [sys.executable, "-m", "pip", "install", "tomli"],
                            check=False,
                            capture_output=True,
                        )
                        import tomli as tomllib  # type: ignore[assignment]
                        toml_parser = tomllib
                    except Exception as install_exc:
                        logger.warning(
                            "tomli installation failed; skipping pyproject.toml discovery "
                            f"({install_exc})"
                        )
                        return

            with open(pyproject_path, 'rb') as f:
                data = toml_parser.load(f)
            
            dependencies = []
            if 'project' in data and 'dependencies' in data['project']:
                dependencies.extend(data['project']['dependencies'])
            
            if 'project' in data and 'optional-dependencies' in data['project']:
                for group_name, group_deps in data['project']['optional-dependencies'].items():
                    dependencies.extend(group_deps)
            
            if 'build-system' in data and 'requires' in data['build-system']:
                dependencies.extend(data['build-system']['requires'])
            
            for dep in dependencies:
                self._parse_dependency_string(dep, 'pyproject.toml')
            
            self._discovery_sources.append('pyproject.toml')
        except Exception as e:
            logger.warning(f"Could not parse pyproject.toml: {e}")
    
    def _discover_from_requirements_txt(self) -> None:
        """Discover dependencies from requirements.txt."""
        requirements_path = self.project_root / 'requirements.txt'
        if not requirements_path.exists():
            return
        
        try:
            with open(requirements_path, 'r', encoding='utf-8') as f:
                for line in f:
                    line = line.strip()
                    if line and not line.startswith('#'):
                        self._parse_dependency_string(line, 'requirements.txt')
            
            self._discovery_sources.append('requirements.txt')
        except Exception as e:
            logger.warning(f"Could not parse requirements.txt: {e}")
    
    def _discover_from_setup_py(self) -> None:
        """Discover dependencies from setup.py."""
        setup_path = self.project_root / 'setup.py'
        if not setup_path.exists():
            return
        
        try:
            with open(setup_path, 'r', encoding='utf-8') as f:
                content = f.read()
            
            install_requires_match = re.search(
                r'install_requires\s*=\s*\[(.*?)\]', 
                content, 
                re.DOTALL
            )
            if install_requires_match:
                deps_str = install_requires_match.group(1)
                deps = re.findall(r'["\']([^"\']+)["\']', deps_str)
                for dep in deps:
                    self._parse_dependency_string(dep, 'setup.py')
            
            self._discovery_sources.append('setup.py')
        except Exception as e:
            logger.warning(f"Could not parse setup.py: {e}")
    
    def _discover_from_custom_config(self) -> None:
        """Discover dependencies from custom configuration files."""
        config_files = [
            'dependency-mappings.json',
            'lazy-dependencies.json',
            'dependencies.json'
        ]
        
        for config_file in config_files:
            config_path = self.project_root / config_file
            if config_path.exists():
                try:
                    with open(config_path, 'r', encoding='utf-8') as f:
                        data = json.load(f)
                    
                    if isinstance(data, dict):
                        for import_name, package_name in data.items():
                            self.discovered_dependencies[import_name] = DependencyInfo(
                                import_name=import_name,
                                package_name=package_name,
                                source=config_file,
                                category='custom'
                            )
                    
                    self._discovery_sources.append(config_file)
                except Exception as e:
                    logger.warning(f"Could not parse {config_file}: {e}")
    
    def _parse_dependency_string(self, dep_str: str, source: str) -> None:
        """Parse a dependency string and extract dependency information."""
        dep_str = re.sub(r'[>=<!=~]+.*', '', dep_str)
        dep_str = re.sub(r'\[.*\]', '', dep_str)
        dep_str = dep_str.strip()
        
        if not dep_str:
            return
        
        import_name = dep_str
        package_name = dep_str
        
        if dep_str in self.COMMON_MAPPINGS:
            package_name = self.COMMON_MAPPINGS[dep_str]
        elif dep_str in self.COMMON_MAPPINGS.values():
            for imp_name, pkg_name in self.COMMON_MAPPINGS.items():
                if pkg_name == dep_str:
                    import_name = imp_name
                    break
        
        self.discovered_dependencies[import_name] = DependencyInfo(
            import_name=import_name,
            package_name=package_name,
            source=source,
            category='discovered'
        )
    
    def _add_common_mappings(self) -> None:
        """Add common mappings that might not be in dependency files."""
        for import_name, package_name in self.COMMON_MAPPINGS.items():
            if import_name not in self.discovered_dependencies:
                self.discovered_dependencies[import_name] = DependencyInfo(
                    import_name=import_name,
                    package_name=package_name,
                    source='common_mappings',
                    category='common'
                )
    
    def get_package_for_import(self, import_name: str) -> Optional[str]:
        """Get package name for a given import name."""
        mapping = self.discover_all_dependencies()
        return mapping.get(import_name)
    
    def get_imports_for_package(self, package_name: str) -> List[str]:
        """Get all possible import names for a package."""
        mapping = self.get_package_import_mapping()
        return mapping.get(package_name, [package_name])
    
    def get_package_import_mapping(self) -> Dict[str, List[str]]:
        """Get mapping of package names to their possible import names."""
        self.discover_all_dependencies()
        
        package_to_imports = {}
        for import_name, dep_info in self.discovered_dependencies.items():
            package_name = dep_info.package_name
            
            if package_name not in package_to_imports:
                package_to_imports[package_name] = [package_name]
            
            if import_name != package_name:
                if import_name not in package_to_imports[package_name]:
                    package_to_imports[package_name].append(import_name)
        
        return package_to_imports
    
    def get_import_package_mapping(self) -> Dict[str, str]:
        """Get mapping of import names to package names."""
        self.discover_all_dependencies()
        return {import_name: dep_info.package_name for import_name, dep_info in self.discovered_dependencies.items()}
    
    def export_to_json(self, file_path: str) -> None:
        """Export discovered dependencies to JSON file."""
        data = {
            'dependencies': {name: info.package_name for name, info in self.discovered_dependencies.items()},
            'sources': self.get_discovery_sources(),
            'total_count': len(self.discovered_dependencies)
        }
        
        with open(file_path, 'w', encoding='utf-8') as f:
            json.dump(data, f, indent=2, ensure_ascii=False)

# Global discovery instance
_discovery: Optional[LazyDiscovery] = None
_discovery_lock = threading.RLock()

def get_lazy_discovery(project_root: Optional[str] = None) -> LazyDiscovery:
    """Get or create global discovery instance."""
    global _discovery
    with _discovery_lock:
        if _discovery is None:
            _discovery = LazyDiscovery(project_root)
        return _discovery

__all__ = ['LazyDiscovery', 'get_lazy_discovery']




"""
Permissive Policy Strategy

Company: eXonware.com
Author: Eng. Muhammad AlShehri
Email: connect@exonware.com

Generation Date: 15-Nov-2025

Permissive policy - allows all packages (default).
"""

from typing import Tuple, List
from ...package.base import APolicyStrategy

class PermissivePolicy(APolicyStrategy):
    """
    Permissive policy strategy - allows all packages.
    
    This is the default policy that doesn't restrict any packages.
    """
    
    def is_allowed(self, package_name: str) -> Tuple[bool, str]:
        """
        Check if package is allowed to be installed.
        
        Args:
            package_name: Package name to check
            
        Returns:
            Tuple of (allowed: bool, reason: str)
        """
        return (True, "Permissive policy allows all packages")
    
    def get_pip_args(self, package_name: str) -> List[str]:
        """
        Get pip arguments based on policy.
        
        Args:
            package_name: Package name
            
        Returns:
            List of pip arguments (empty for permissive policy)
        """
        return []




"""
Interactive Installation Mixin

Company: eXonware.com
Author: Eng. Muhammad AlShehri
Email: connect@exonware.com

Generation Date: 15-Nov-2025

Mixin for interactive user prompts during installation.
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .lazy_installer import LazyInstaller

class InteractiveInstallMixin:
    """Mixin for interactive user prompts during installation."""
    
    def _ask_user_permission(self, package_name: str, module_name: str) -> bool:
        """Ask user for permission to install a package."""
        if self._auto_approve_all:  # type: ignore[attr-defined]
            return True
        
        print(f"\n{'='*60}")
        print(f"Lazy Installation Active - {self._package_name}")  # type: ignore[attr-defined]
        print(f"{'='*60}")
        print(f"Package: {package_name}")
        print(f"Module:  {module_name}")
        print(f"{'='*60}")
        print(f"\nThe module '{module_name}' is not installed.")
        print(f"Would you like to install '{package_name}'?")
        print(f"\nOptions:")
        print(f"  [Y] Yes - Install this package")
        print(f"  [N] No  - Skip this package")
        print(f"  [A] All - Install this and all future packages without asking")
        print(f"  [Q] Quit - Cancel and raise ImportError")
        print(f"{'='*60}")
        
        while True:
            try:
                choice = input("Your choice [Y/N/A/Q]: ").strip().upper()
                
                if choice in ('Y', 'YES', ''):
                    return True
                elif choice in ('N', 'NO'):
                    return False
                elif choice in ('A', 'ALL'):
                    self._auto_approve_all = True  # type: ignore[attr-defined]
                    return True
                elif choice in ('Q', 'QUIT'):
                    raise KeyboardInterrupt("User cancelled installation")
                else:
                    print(f"Invalid choice '{choice}'. Please enter Y, N, A, or Q.")
            except (EOFError, KeyboardInterrupt):
                print("\n❌ Installation cancelled by user")
                return False




"""
Install Result and Status

Company: eXonware.com
Author: Eng. Muhammad AlShehri
Email: connect@exonware.com

Generation Date: 15-Nov-2025

Re-export installation result types from defs.py for backward compatibility.
"""

# Re-export from defs.py
from ...defs import InstallStatus, InstallResult

__all__ = ['InstallStatus', 'InstallResult']




"""
Common Caching Strategies - Shared by modules and packages.

Company: eXonware.com
Author: Eng. Muhammad AlShehri
Email: connect@exonware.com

Generation Date: 15-Nov-2025

Generic caching strategies that work with ANY data type.
"""

from .caching_dict import DictCache
from .caching_lru import LRUCache
from .caching_lfu import LFUCache
from .caching_ttl import TTLCache
from .caching_multitier import MultiTierCacheStrategy
from .caching_installation import InstallationCacheWrapper

__all__ = [
    'DictCache',
    'LRUCache',
    'LFUCache',
    'TTLCache',
    'MultiTierCacheStrategy',
    'InstallationCacheWrapper',
]




"""
Installer Registry

Company: eXonware.com
Author: Eng. Muhammad AlShehri
Email: connect@exonware.com

Generation Date: 15-Nov-2025

Registry to manage separate lazy installer instances per package.
"""

import threading
from typing import Dict, TYPE_CHECKING

if TYPE_CHECKING:
    from .lazy_installer import LazyInstaller

class LazyInstallerRegistry:
    """Registry to manage separate lazy installer instances per package."""
    _instances: Dict[str, 'LazyInstaller'] = {}
    _lock = threading.RLock()
    
    @classmethod
    def get_instance(cls, package_name: str = 'default') -> 'LazyInstaller':
        """
        Get or create a lazy installer instance for a package.
        
        Args:
            package_name: Package name for isolation
            
        Returns:
            LazyInstaller instance for the package
        """
        with cls._lock:
            if package_name not in cls._instances:
                # Lazy import to avoid circular dependency
                from .lazy_installer import LazyInstaller
                cls._instances[package_name] = LazyInstaller(package_name)
            return cls._instances[package_name]
    
    @classmethod
    def get_all_instances(cls) -> Dict[str, 'LazyInstaller']:
        """
        Get all lazy installer instances.
        
        Returns:
            Dict mapping package_name -> LazyInstaller
        """
        with cls._lock:
            return cls._instances.copy()

__all__ = ['LazyInstallerRegistry']




"""
Module Operations Facade

Main facade: XWModuleHelper extends AModuleHelper
Provides concrete implementation for all module operations.
Uses strategy pattern for caching, helper, and manager strategies.
"""

import sys
import importlib
import importlib.util
from typing import Optional, Dict
from types import ModuleType

from .base import AModuleHelper, APackageHelper
# Lazy import to avoid circular dependency
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..package.facade import XWPackageHelper

# Import from importer_engine for hook and module operations
from . import importer_engine

# Import strategy interfaces
from ..contracts import ICachingStrategy, IModuleHelperStrategy, IModuleManagerStrategy

# Import default strategies
from ..common.strategies import LRUCache
from .strategies import LazyHelper, AdvancedManager
from ..package.services.strategy_registry import StrategyRegistry

class XWModuleHelper(AModuleHelper):
    """
    Concrete implementation of AModuleHelper.
    
    Provides simple, clean API for working with modules (what you import).
    Uses XWPackageHelper for package operations and DependencyMapper for module-to-package mapping.
    """
    
    def __init__(
        self,
        package_name: str = 'default',
        package_helper: Optional[APackageHelper] = None,
        *,
        # Strategy injection
        caching_strategy: Optional[ICachingStrategy] = None,
        helper_strategy: Optional[IModuleHelperStrategy] = None,
        manager_strategy: Optional[IModuleManagerStrategy] = None
    ):
        """
        Initialize XW module helper.
        
        Args:
            package_name: Package name for isolation (defaults to 'default')
            package_helper: Optional package helper instance. If None, creates XWPackageHelper.
            caching_strategy: Optional caching strategy. If None, uses LRUCache.
            helper_strategy: Optional helper strategy. If None, uses LazyHelper.
            manager_strategy: Optional manager strategy. If None, uses AdvancedManager.
        """
        if package_helper is None:
            # Lazy import to avoid circular dependency
            from ..package.facade import XWPackageHelper
            package_helper = XWPackageHelper(package_name)
        super().__init__(package_name, package_helper)
        self._package_name = package_name
        self._package_helper = package_helper
        
        # Check registry for stored strategies, otherwise use defaults
        if caching_strategy is None:
            caching_strategy = StrategyRegistry.get_module_strategy(package_name, 'caching')
            if caching_strategy is None:
                caching_strategy = LRUCache(max_size=1000)
        if helper_strategy is None:
            helper_strategy = StrategyRegistry.get_module_strategy(package_name, 'helper')
            if helper_strategy is None:
                helper_strategy = LazyHelper()
        if manager_strategy is None:
            manager_strategy = StrategyRegistry.get_module_strategy(package_name, 'manager')
            if manager_strategy is None:
                manager_strategy = AdvancedManager(
                    package_name,
                    package_helper,
                    caching_strategy,
                    helper_strategy
                )
        
        # Store strategies
        self._caching = caching_strategy
        self._helper = helper_strategy
        self._manager = manager_strategy
    
    def _check_module_importability(self, module_name: str) -> bool:
        """
        Check if module is importable.
        
        Uses importlib.util.find_spec to check if module can be imported.
        
        Args:
            module_name: Module name to check
            
        Returns:
            True if importable, False otherwise
        """
        try:
            spec = importlib.util.find_spec(module_name)
            return spec is not None
        except (ValueError, AttributeError, ImportError):
            return False
    
    def _import_module(self, module_name: str) -> ModuleType:
        """
        Import a module.
        
        Uses importlib.import_module to import the module.
        
        Args:
            module_name: Module name to import
            
        Returns:
            Imported module
            
        Raises:
            ImportError: If module cannot be imported
        """
        return importlib.import_module(module_name)
    
    def _invalidate_import_caches(self) -> None:
        """
        Invalidate import caches.
        
        Uses importlib.invalidate_caches() to clear Python's import caches.
        """
        importlib.invalidate_caches()
        sys.path_importer_cache.clear()
    
    def _create_package_helper(self) -> APackageHelper:
        """
        Create a package helper instance.
        
        Returns:
            XWPackageHelper instance
        """
        from ..package.facade import XWPackageHelper
        return XWPackageHelper(self._package_name)
    
    # Strategy swapping methods
    def swap_cache_strategy(self, new_strategy: ICachingStrategy) -> None:
        """
        Swap cache strategy at runtime.
        
        Args:
            new_strategy: New caching strategy to use
        """
        self._caching = new_strategy
        # Update manager if it uses caching
        if hasattr(self._manager, '_caching'):
            self._manager._caching = new_strategy
    
    def swap_helper_strategy(self, new_strategy: IModuleHelperStrategy) -> None:
        """
        Swap helper/engine strategy at runtime.
        
        Args:
            new_strategy: New helper strategy to use
        """
        self._helper = new_strategy
        # Update manager if it uses helper
        if hasattr(self._manager, '_helper'):
            self._manager._helper = new_strategy
    
    def swap_manager_strategy(self, new_strategy: IModuleManagerStrategy) -> None:
        """
        Swap manager strategy at runtime.
        
        Args:
            new_strategy: New manager strategy to use
        """
        self._manager = new_strategy
    
    # Abstract methods from AModule that need implementation
    def install_hook(self) -> None:
        """
        Install the import hook into sys.meta_path.
        
        Delegates to manager strategy.
        """
        self._manager.install_hook()
    
    def uninstall_hook(self) -> None:
        """
        Uninstall the import hook from sys.meta_path.
        
        Delegates to manager strategy.
        """
        self._manager.uninstall_hook()
    
    def is_hook_installed(self) -> bool:
        """
        Check if import hook is installed.
        
        Uses importer_engine.is_import_hook_installed() to check hook status.
        
        Returns:
            True if hook is installed, False otherwise
        """
        return importer_engine.is_import_hook_installed(self._package_name)
    
    def handle_import_error(self, module_name: str) -> Optional[ModuleType]:
        """
        Handle ImportError by attempting to install and re-import.
        
        Delegates to manager strategy.
        
        Args:
            module_name: Name of module that failed to import
            
        Returns:
            Imported module if successful, None otherwise
        """
        return self._manager.handle_import_error(module_name)
    
    def load_module(self, module_path: str) -> ModuleType:
        """
        Load a module lazily.
        
        Delegates to manager strategy.
        
        Args:
            module_path: Full module path to load
            
        Returns:
            Loaded module
        """
        return self._manager.load_module(module_path)
    
    def unload_module(self, module_path: str) -> None:
        """
        Unload a module from cache.
        
        Delegates to manager strategy.
        
        Args:
            module_path: Module path to unload
        """
        self._manager.unload_module(module_path)




"""
#exonware/xwlazy/src/exonware/xwlazy/contracts.py

Company: eXonware.com
Author: Eng. Muhammad AlShehri
Email: connect@exonware.com

Generation Date: 10-Oct-2025

Contracts for Lazy Loading System

This module defines all interfaces, enums, and protocols for the lazy loading
system following GUIDE_ARCH.md structure.
"""

from typing import Protocol, Dict, List, Optional, Any, Tuple, runtime_checkable
from types import ModuleType

# Import enums and dataclasses from defs.py
from .defs import (
    LazyLoadMode,
    LazyInstallMode,
    PathType,
    DependencyInfo,
    LazyModeConfig,
)

# =============================================================================
# PROTOCOLS / INTERFACES (Following GUIDE_DEV.md - Use IClass naming)
# =============================================================================

# NOTE: Old interfaces have been merged into three unified interfaces:
# - IPackageHelper (package operations: discovery, installation, caching, config, manifest, mapping)
# - IModuleHelper (module operations: installation, hooks, finding, interception, loading, importing, registry, bytecode)
# - IRuntime (runtime services: state, learning, selection, metrics, monitoring, caching, registry)

class IPackageHelper(Protocol):
    """
    Unified interface for package operations.
    
    Merges functionality from IPackageDiscovery, IPackageInstaller, IPackageCache,
    IConfigManager, IManifestLoader, and IDependencyMapper.
    Provides simple, clean API for working with packages (what you pip install).
    Packages are what you install via pip: pip install pymongo, pip install msgpack, etc.
    
    This interface combines:
    - Package discovery (mapping import names to package names)
    - Package installation (installing/uninstalling packages)
    - Package caching (caching installation status and metadata)
    - Configuration management (per-package lazy installation configuration)
    - Manifest loading (loading and caching dependency manifests)
    - Dependency mapping (mapping import names to package names)
    """
    
    # ========================================================================
    # Installation Status (from IPackageInstaller + IPackageHelper)
    # ========================================================================
    
    def installed(self, package_name: str) -> bool:
        """
        Check if a package is installed.
        
        Uses cache first to avoid expensive operations.
        Checks persistent cache, then in-memory cache, then importability.
        
        Args:
            package_name: Package name to check (e.g., 'pymongo', 'msgpack')
            
        Returns:
            True if package is installed, False otherwise
        """
        ...
    
    def uninstalled(self, package_name: str) -> bool:
        """
        Check if a package is uninstalled.
        
        Uses cache first to avoid expensive operations.
        Checks persistent cache, then in-memory cache, then importability.
        
        Args:
            package_name: Package name to check (e.g., 'pymongo', 'msgpack')
            
        Returns:
            True if package is uninstalled, False otherwise
        """
        ...
    
    # ========================================================================
    # Package Installation (from IPackageInstaller + IPackageHelper)
    # ========================================================================
    
    def install(self, *package_names: str) -> None:
        """
        Install one or more packages using pip.
        
        Skips packages that are already installed (using cache).
        Only installs unique packages to avoid duplicate operations.
        Updates cache after successful installation.
        
        Args:
            *package_names: One or more package names to install (e.g., 'pymongo', 'msgpack')
            
        Raises:
            subprocess.CalledProcessError: If installation fails
        """
        ...
    
    def install_package(self, package_name: str, module_name: Optional[str] = None) -> bool:
        """
        Install a single package (from IPackageInstaller).
        
        Args:
            package_name: Name of package to install
            module_name: Optional name of module being imported (for interactive mode)
            
        Returns:
            True if installation successful, False otherwise
        """
        ...
    
    def install_and_import(self, module_name: str, package_name: Optional[str] = None) -> Tuple[Optional[ModuleType], bool]:
        """
        Install package and import module (from IPackageInstaller).
        
        Args:
            module_name: Name of module to import
            package_name: Optional package name if different from module name
            
        Returns:
            Tuple of (module_object, success_flag)
        """
        ...
    
    def uninstall(self, *package_names: str) -> None:
        """
        Uninstall one or more packages using pip.
        
        Skips packages that are already uninstalled (using cache).
        Only uninstalls unique packages to avoid duplicate operations.
        Updates cache after successful uninstallation.
        
        Args:
            *package_names: One or more package names to uninstall (e.g., 'pymongo', 'msgpack')
            
        Raises:
            subprocess.CalledProcessError: If uninstallation fails
        """
        ...
    
    # ========================================================================
    # Package Discovery (from IPackageDiscovery)
    # ========================================================================
    
    def discover_all_dependencies(self) -> Dict[str, str]:
        """
        Discover all dependencies from all available sources.
        
        Discovers from pyproject.toml, requirements.txt, setup.py, etc.
        
        Returns:
            Dict mapping import_name -> package_name
        """
        ...
    
    def get_package_for_import(self, import_name: str) -> Optional[str]:
        """
        Get package name for a given import name.
        
        Args:
            import_name: Import name (e.g., 'cv2', 'PIL')
            
        Returns:
            Package name (e.g., 'opencv-python', 'Pillow') or None
        """
        ...
    
    def get_imports_for_package(self, package_name: str) -> List[str]:
        """
        Get all possible import names for a package.
        
        Args:
            package_name: Package name (e.g., 'opencv-python')
            
        Returns:
            List of import names (e.g., ['opencv-python', 'cv2'])
        """
        ...
    
    # ========================================================================
    # Package Caching (from IPackageCache)
    # ========================================================================
    
    def get_cached(self, key: str) -> Optional[Any]:
        """
        Get cached value.
        
        Args:
            key: Cache key
            
        Returns:
            Cached value or None if not found
        """
        ...
    
    def set_cached(self, key: str, value: Any) -> None:
        """
        Set cached value.
        
        Args:
            key: Cache key
            value: Value to cache
        """
        ...
    
    def clear_cache(self) -> None:
        """Clear all cached values."""
        ...
    
    def is_cache_valid(self, key: str) -> bool:
        """
        Check if cache entry is still valid.
        
        Args:
            key: Cache key
            
        Returns:
            True if valid, False otherwise
        """
        ...
    
    # ========================================================================
    # Configuration Management (from IConfigManager)
    # ========================================================================
    
    def is_enabled(self, package_name: str) -> bool:
        """
        Check if lazy install is enabled for a package.
        
        Args:
            package_name: Package name
            
        Returns:
            True if enabled, False otherwise
        """
        ...
    
    def get_mode(self, package_name: str) -> str:
        """
        Get installation mode for a package.
        
        Args:
            package_name: Package name
            
        Returns:
            Mode string
        """
        ...
    
    def get_load_mode(self, package_name: str) -> LazyLoadMode:
        """
        Get load mode for a package.
        
        Args:
            package_name: Package name
            
        Returns:
            LazyLoadMode enum
        """
        ...
    
    def get_install_mode(self, package_name: str) -> LazyInstallMode:
        """
        Get install mode for a package.
        
        Args:
            package_name: Package name
            
        Returns:
            LazyInstallMode enum
        """
        ...
    
    def get_mode_config(self, package_name: str) -> Optional[LazyModeConfig]:
        """
        Get full mode configuration for a package.
        
        Args:
            package_name: Package name
            
        Returns:
            LazyModeConfig or None
        """
        ...
    
    # ========================================================================
    # Manifest Loading (from IManifestLoader)
    # ========================================================================
    
    def get_manifest_signature(self, package_name: str) -> Optional[Tuple[str, float, float]]:
        """
        Get manifest file signature (path, mtime, size).
        
        Args:
            package_name: Package name
            
        Returns:
            Tuple of (path, mtime, size) or None
        """
        ...
    
    def get_shared_dependencies(self, package_name: str, signature: Optional[Tuple[str, float, float]] = None) -> Dict[str, str]:
        """
        Get shared dependencies from manifest.
        
        Args:
            package_name: Package name
            signature: Optional signature for cache validation
            
        Returns:
            Dict mapping import_name -> package_name
        """
        ...
    
    def get_watched_prefixes(self, package_name: str) -> Tuple[str, ...]:
        """
        Get watched prefixes from manifest.
        
        Args:
            package_name: Package name
            
        Returns:
            Tuple of watched prefixes
        """
        ...
    
    # ========================================================================
    # Dependency Mapping (from IDependencyMapper)
    # ========================================================================
    
    def get_package_name(self, import_name: str) -> Optional[str]:
        """
        Get package name for an import name (from IDependencyMapper).
        
        Note: This is similar to get_package_for_import() but uses different naming.
        Both methods serve the same purpose - mapping import names to package names.
        
        Args:
            import_name: Import name (e.g., 'cv2', 'msgpack')
            
        Returns:
            Package name (e.g., 'opencv-python', 'msgpack') or None
        """
        ...
    
    def get_import_names(self, package_name: str) -> List[str]:
        """
        Get all import names for a package (from IDependencyMapper).
        
        Note: This is similar to get_imports_for_package() but uses different naming.
        Both methods serve the same purpose - mapping package names to import names.
        
        Args:
            package_name: Package name
            
        Returns:
            List of import names
        """
        ...
    
    def is_stdlib_or_builtin(self, import_name: str) -> bool:
        """
        Check if import name is stdlib or builtin.
        
        Args:
            import_name: Import name to check
            
        Returns:
            True if stdlib/builtin, False otherwise
        """
        ...

class IModuleHelper(Protocol):
    """
    Unified interface for module operations.
    
    Merges functionality from IModuleInstaller, IImportHook, IMetaPathFinder,
    IImportInterceptor, ILazyLoader, ILazyImporter, IWatchedRegistry, and IBytecodeCache.
    Provides simple, clean API for working with modules (what you import).
    Modules are what you use in Python code: import bson, import msgpack, etc.
    
    This interface combines:
    - Module installation (installing and importing modules)
    - Import hooks (intercepting import failures)
    - Meta path finding (sys.meta_path hook for lazy installation)
    - Import interception (high-level import interception)
    - Lazy loading (deferred module loading)
    - Lazy importing (lazy module loading with strategies)
    - Watched registry (tracking watched module prefixes)
    - Bytecode caching (caching compiled Python bytecode)
    """
    
    # ========================================================================
    # Module Operations (from IModuleHelper - existing)
    # ========================================================================
    
    def to_package(self, module_name: str) -> Optional[str]:
        """
        Map module name to package name.
        
        Args:
            module_name: Module name (e.g., 'bson', 'msgpack')
            
        Returns:
            Package name (e.g., 'pymongo', 'msgpack') or None if not found
        """
        ...
    
    def installed(self, module_name: str) -> bool:
        """
        Check if a module is installed.
        
        Uses cache first to avoid expensive operations.
        Maps module to package and checks if package is installed.
        
        Args:
            module_name: Module name to check (e.g., 'bson')
            
        Returns:
            True if module is installed, False otherwise
        """
        ...
    
    def uninstalled(self, module_name: str) -> bool:
        """
        Check if a module is uninstalled.
        
        Uses cache first to avoid expensive operations.
        Maps module to package and checks if package is uninstalled.
        
        Args:
            module_name: Module name to check (e.g., 'bson')
            
        Returns:
            True if module is uninstalled, False otherwise
        """
        ...
    
    def install(self, *module_names: str) -> None:
        """
        Install one or more modules by mapping to packages first.
        
        First deduplicates modules, then maps to packages, then installs packages.
        Skips modules that are already installed (using cache).
        
        Args:
            *module_names: One or more module names to install (e.g., 'bson', 'msgpack')
            
        Raises:
            subprocess.CalledProcessError: If installation fails
        """
        ...
    
    def uninstall(self, *module_names: str) -> None:
        """
        Uninstall one or more modules by mapping to packages first.
        
        First deduplicates modules, then maps to packages, then uninstalls packages.
        Skips modules that are already uninstalled (using cache).
        
        Args:
            *module_names: One or more module names to uninstall (e.g., 'bson', 'msgpack')
            
        Raises:
            subprocess.CalledProcessError: If uninstallation fails
        """
        ...
    
    def load(self, *module_names: str) -> List[ModuleType]:
        """
        Load one or more modules into memory.
        
        Imports modules and returns them. Uses lazy loading if enabled.
        
        Args:
            *module_names: One or more module names to load (e.g., 'bson', 'msgpack')
            
        Returns:
            List of loaded module objects
            
        Raises:
            ImportError: If module cannot be loaded
        """
        ...
    
    def unload(self, *module_names: str) -> None:
        """
        Unload one or more modules from memory.
        
        Removes modules from sys.modules and clears caches.
        Useful for freeing memory or forcing reload.
        
        Args:
            *module_names: One or more module names to unload (e.g., 'bson', 'msgpack')
        """
        ...
    
    # ========================================================================
    # Module Installation (from IModuleInstaller)
    # ========================================================================
    
    def install_and_import(self, module_name: str, package_name: Optional[str] = None) -> Tuple[Optional[ModuleType], bool]:
        """
        Install package and import module (from IModuleInstaller).
        
        CONTRACT:
        1. Check persistent cache FIRST - if installed, import directly
        2. Check importability - if importable, import directly
        3. If not importable, remove finders from sys.meta_path
        4. Install package via pip
        5. Restore finders
        6. Import module
        7. Mark in persistent cache
        8. Return (module, success)
        
        Args:
            module_name: Name of module to import
            package_name: Optional package name (if different from module)
            
        Returns:
            Tuple of (module_object, success_flag)
        """
        ...
    
    def is_package_installed(self, package_name: str) -> bool:
        """
        Check if package is installed (from IModuleInstaller).
        
        CONTRACT:
        - Check persistent cache FIRST
        - Check in-memory cache second
        - Check importability last
        
        Args:
            package_name: Package name to check
            
        Returns:
            True if installed, False otherwise
        """
        ...
    
    def mark_installed(self, package_name: str, version: Optional[str] = None) -> None:
        """
        Mark package as installed in persistent cache (from IModuleInstaller).
        
        Args:
            package_name: Package name
            version: Optional version string
        """
        ...
    
    # ========================================================================
    # Import Hooks (from IImportHook)
    # ========================================================================
    
    def install_hook(self) -> None:
        """Install the import hook into sys.meta_path."""
        ...
    
    def uninstall_hook(self) -> None:
        """Uninstall the import hook from sys.meta_path."""
        ...
    
    def is_hook_installed(self) -> bool:
        """
        Check if hook is installed.
        
        Returns:
            True if hook is in sys.meta_path, False otherwise
        """
        ...
    
    def handle_import_error(self, module_name: str) -> Optional[Any]:
        """
        Handle ImportError by attempting to install and re-import (from IImportHook).
        
        Args:
            module_name: Name of module that failed to import
            
        Returns:
            Imported module if successful, None otherwise
        """
        ...
    
    # ========================================================================
    # Meta Path Finding (from IMetaPathFinder)
    # ========================================================================
    
    def find_spec(self, fullname: str, path: Optional[str] = None, target=None) -> Optional[Any]:
        """
        Find module spec - intercepts imports to enable lazy installation (from IMetaPathFinder).
        
        CONTRACT:
        - If module is already in sys.modules: return None (let Python use it)
        - If module starts with '_': return None (skip C extensions/internal modules)
        - If lazy install is disabled: only intercept watched modules
        - If lazy install is enabled: intercept ALL missing imports
        - If module is in persistent cache as installed: return None (let Python import normally)
        - If module is missing and lazy install enabled: install it, then return None
        
        Args:
            fullname: Full module name (e.g., 'msgpack', 'yaml._yaml')
            path: Optional path (for package imports)
            target: Optional target module
            
        Returns:
            ModuleSpec if intercepted, None to let Python handle normally
        """
        ...
    
    def should_intercept(self, fullname: str) -> bool:
        """
        Determine if a module should be intercepted (from IMetaPathFinder).
        
        CONTRACT:
        - Return False if module is in sys.modules
        - Return False if module starts with '_' (C extension/internal)
        - Return False if lazy install disabled and module not watched
        - Return True if lazy install enabled and module not in cache
        
        Args:
            fullname: Full module name
            
        Returns:
            True if should intercept, False otherwise
        """
        ...
    
    def is_module_installed(self, fullname: str) -> bool:
        """
        Check if module is already installed (in cache or importable) (from IMetaPathFinder).
        
        CONTRACT:
        - Check persistent cache FIRST (fastest)
        - Check in-memory cache second
        - Check importability last (slowest)
        
        Args:
            fullname: Full module name
            
        Returns:
            True if installed, False otherwise
        """
        ...
    
    # ========================================================================
    # Import Interception (from IImportInterceptor)
    # ========================================================================
    
    def intercept_missing_import(self, module_name: str) -> Optional[ModuleType]:
        """
        Intercept a missing import and attempt to install it (from IImportInterceptor).
        
        CONTRACT:
        1. Check if module is already installed (persistent cache)
        2. If installed, return None (let Python import normally)
        3. If not installed and lazy install enabled:
           a. Install package
           b. Import module
           c. Mark in persistent cache
           d. Return module
        4. If lazy install disabled, return None
        
        Args:
            module_name: Name of module that failed to import
            
        Returns:
            Imported module if successful, None otherwise
        """
        ...
    
    def should_intercept_module(self, module_name: str) -> bool:
        """
        Determine if a module should be intercepted (from IImportInterceptor).
        
        CONTRACT:
        - Return False if module starts with '_' (C extension/internal)
        - Return False if module is stdlib/builtin
        - Return False if module is in deny list
        - Return False if lazy install disabled and module not watched
        - Return True if lazy install enabled
        
        Args:
            module_name: Module name to check
            
        Returns:
            True if should intercept, False otherwise
        """
        ...
    
    def prevent_recursion(self, module_name: str) -> bool:
        """
        Check if we should prevent recursion for this module (from IImportInterceptor).
        
        CONTRACT:
        - Return True if installation is in progress
        - Return True if import is in progress for this module
        - Return False otherwise
        
        Args:
            module_name: Module name to check
            
        Returns:
            True if should prevent recursion, False otherwise
        """
        ...
    
    # ========================================================================
    # Lazy Loading (from ILazyLoader)
    # ========================================================================
    
    def load_module(self, module_path: str) -> ModuleType:
        """
        Load a module lazily (from ILazyLoader).
        
        Args:
            module_path: Full module path to load
            
        Returns:
            Loaded module
        """
        ...
    
    def is_loaded(self, module_path: str) -> bool:
        """
        Check if module is already loaded (from ILazyLoader).
        
        Args:
            module_path: Module path to check
            
        Returns:
            True if loaded, False otherwise
        """
        ...
    
    def unload_module(self, module_path: str) -> None:
        """
        Unload a module from cache (from ILazyLoader).
        
        Args:
            module_path: Module path to unload
        """
        ...
    
    # ========================================================================
    # Lazy Importing (from ILazyImporter)
    # ========================================================================
    
    def import_module(self, module_name: str, package_name: Optional[str] = None) -> Any:
        """
        Import a module with lazy loading (from ILazyImporter).
        
        Args:
            module_name: Module name to import
            package_name: Optional package name
            
        Returns:
            Imported module
        """
        ...
    
    def enable_lazy_loading(self, load_mode: LazyLoadMode) -> None:
        """
        Enable lazy loading with a mode (from ILazyImporter).
        
        Args:
            load_mode: LazyLoadMode to use
        """
        ...
    
    def disable_lazy_loading(self) -> None:
        """Disable lazy loading (from ILazyImporter)."""
        ...
    
    def is_lazy_loading_enabled(self) -> bool:
        """
        Check if lazy loading is enabled (from ILazyImporter).
        
        Returns:
            True if enabled, False otherwise
        """
        ...
    
    # ========================================================================
    # Watched Registry (from IWatchedRegistry)
    # ========================================================================
    
    def has_root(self, root_name: str) -> bool:
        """
        Check if a root module name is being watched (from IWatchedRegistry).
        
        Args:
            root_name: Root module name (e.g., 'msgpack', 'yaml')
            
        Returns:
            True if watched, False otherwise
        """
        ...
    
    def get_matching_prefixes(self, fullname: str) -> Tuple[str, ...]:
        """
        Get all watched prefixes that match a module name (from IWatchedRegistry).
        
        Args:
            fullname: Full module name
            
        Returns:
            Tuple of matching prefixes
        """
        ...
    
    def is_prefix_owned_by(self, prefix: str, package_name: str) -> bool:
        """
        Check if a prefix is owned by a package (from IWatchedRegistry).
        
        Args:
            prefix: Prefix to check
            package_name: Package name
            
        Returns:
            True if owned, False otherwise
        """
        ...
    
    def is_watched_registry_empty(self) -> bool:
        """
        Check if registry is empty (from IWatchedRegistry).
        
        Returns:
            True if empty, False otherwise
        """
        ...
    
    # ========================================================================
    # Bytecode Caching (from IBytecodeCache)
    # ========================================================================
    
    def get_bytecode(self, module_path: str, source_code: str) -> Optional[bytes]:
        """
        Get cached bytecode for module (from IBytecodeCache).
        
        Args:
            module_path: Module path
            source_code: Source code
            
        Returns:
            Cached bytecode or None
        """
        ...
    
    def cache_bytecode(self, module_path: str, source_code: str, bytecode: bytes) -> None:
        """
        Cache bytecode for module (from IBytecodeCache).
        
        Args:
            module_path: Module path
            source_code: Source code
            bytecode: Compiled bytecode
        """
        ...
    
    def clear_bytecode_cache(self) -> None:
        """Clear bytecode cache (from IBytecodeCache)."""
        ...

class IRuntime(Protocol):
    """
    Unified interface for runtime services.
    
    Merges functionality from IStateManager, IAdaptiveLearner, IIntelligentSelector,
    IMetricsCollector, IPerformanceMonitor, IMultiTierCache, and IRegistry.
    Provides runtime services for state management, learning, monitoring, and caching.
    
    This interface combines:
    - State management (persistent state for lazy installation)
    - Adaptive learning (learning import patterns and optimizing loading)
    - Intelligent selection (selecting optimal modes based on load characteristics)
    - Metrics collection (collecting and aggregating performance metrics)
    - Performance monitoring (monitoring lazy loading performance)
    - Multi-tier caching (L1/L2/L3 caching)
    - Registry (managing instances by key)
    """
    
    # ========================================================================
    # State Management (from IStateManager)
    # ========================================================================
    
    def get_manual_state(self) -> Optional[bool]:
        """
        Get manual state override (from IStateManager).
        
        Returns:
            True/False if manually set, None otherwise
        """
        ...
    
    def set_manual_state(self, value: Optional[bool]) -> None:
        """
        Set manual state override (from IStateManager).
        
        Args:
            value: True/False to set, None to clear
        """
        ...
    
    def get_cached_auto_state(self) -> Optional[bool]:
        """
        Get cached auto-detected state (from IStateManager).
        
        Returns:
            True/False if cached, None otherwise
        """
        ...
    
    def set_auto_state(self, value: Optional[bool]) -> None:
        """
        Set cached auto-detected state (from IStateManager).
        
        Args:
            value: True/False to cache, None to clear
        """
        ...
    
    # ========================================================================
    # Adaptive Learning (from IAdaptiveLearner)
    # ========================================================================
    
    def record_import(self, module_name: str, import_time: float) -> None:
        """
        Record an import event (from IAdaptiveLearner).
        
        Args:
            module_name: Module name that was imported
            import_time: Time taken to import (seconds)
        """
        ...
    
    def predict_next_imports(self, current_module: str, count: int = 3) -> List[str]:
        """
        Predict next likely imports based on patterns (from IAdaptiveLearner).
        
        Args:
            current_module: Current module name
            count: Number of predictions to return
            
        Returns:
            List of predicted module names
        """
        ...
    
    def get_module_score(self, module_name: str) -> float:
        """
        Get priority score for a module (from IAdaptiveLearner).
        
        Args:
            module_name: Module name
            
        Returns:
            Priority score (higher = more important)
        """
        ...
    
    # ========================================================================
    # Intelligent Selection (from IIntelligentSelector)
    # ========================================================================
    
    def detect_load_level(
        self,
        module_count: int = 0,
        total_import_time: float = 0.0,
        import_count: int = 0,
        memory_usage_mb: float = 0.0
    ) -> Any:
        """
        Detect current load level (from IIntelligentSelector).
        
        Args:
            module_count: Number of modules loaded
            total_import_time: Total import time (seconds)
            import_count: Number of imports
            memory_usage_mb: Memory usage in MB
            
        Returns:
            LoadLevel enum
        """
        ...
    
    def get_optimal_mode(self, load_level: Any) -> Tuple[LazyLoadMode, LazyInstallMode]:
        """
        Get optimal mode for a load level (from IIntelligentSelector).
        
        Args:
            load_level: LoadLevel enum
            
        Returns:
            Tuple of (LazyLoadMode, LazyInstallMode)
        """
        ...
    
    def update_mode_map(self, mode_map: Dict[Any, Tuple[LazyLoadMode, LazyInstallMode]]) -> None:
        """
        Update mode mapping with benchmark results (from IIntelligentSelector).
        
        Args:
            mode_map: New mode mapping
        """
        ...
    
    # ========================================================================
    # Metrics Collection (from IMetricsCollector)
    # ========================================================================
    
    def record_metric(self, name: str, value: float, timestamp: Optional[Any] = None) -> None:
        """
        Record a metric value (from IMetricsCollector).
        
        Args:
            name: Metric name
            value: Metric value
            timestamp: Optional timestamp
        """
        ...
    
    def get_metric_stats(self, name: str) -> Dict[str, Any]:
        """
        Get statistics for a metric (from IMetricsCollector).
        
        Args:
            name: Metric name
            
        Returns:
            Dict with count, total, average, min, max
        """
        ...
    
    def get_all_stats(self) -> Dict[str, Dict[str, Any]]:
        """
        Get statistics for all metrics (from IMetricsCollector).
        
        Returns:
            Dict mapping metric name -> stats
        """
        ...
    
    def clear_metrics(self) -> None:
        """Clear all collected metrics (from IMetricsCollector)."""
        ...
    
    # ========================================================================
    # Performance Monitoring (from IPerformanceMonitor)
    # ========================================================================
    
    def record_load_time(self, module: str, load_time: float) -> None:
        """
        Record module load time (from IPerformanceMonitor).
        
        Args:
            module: Module name
            load_time: Load time in seconds
        """
        ...
    
    def record_access(self, module: str) -> None:
        """
        Record module access (from IPerformanceMonitor).
        
        Args:
            module: Module name
        """
        ...
    
    def get_performance_stats(self) -> Dict[str, Any]:
        """
        Get performance statistics (from IPerformanceMonitor).
        
        Returns:
            Dict with load_times, access_counts, memory_usage
        """
        ...
    
    # ========================================================================
    # Multi-Tier Cache (from IMultiTierCache)
    # ========================================================================
    
    def get_cached_value(self, key: str) -> Optional[Any]:
        """
        Get value from cache (L1 -> L2 -> L3) (from IMultiTierCache).
        
        Args:
            key: Cache key
            
        Returns:
            Cached value or None
        """
        ...
    
    def set_cached_value(self, key: str, value: Any) -> None:
        """
        Set value in cache (L1 and L2) (from IMultiTierCache).
        
        Args:
            key: Cache key
            value: Value to cache
        """
        ...
    
    def clear_multi_tier_cache(self) -> None:
        """Clear all cache tiers (from IMultiTierCache)."""
        ...
    
    def shutdown_cache(self) -> None:
        """Shutdown cache (flush L2, cleanup threads) (from IMultiTierCache)."""
        ...
    
    # ========================================================================
    # Registry (from IRegistry)
    # ========================================================================
    
    def get_instance(self, key: str) -> Any:
        """
        Get instance by key (from IRegistry).
        
        Args:
            key: Registry key
            
        Returns:
            Registered instance
        """
        ...
    
    def register(self, key: str, instance: Any) -> None:
        """
        Register an instance (from IRegistry).
        
        Args:
            key: Registry key
            instance: Instance to register
        """
        ...
    
    def unregister(self, key: str) -> None:
        """
        Unregister an instance (from IRegistry).
        
        Args:
            key: Registry key
        """
        ...
    
    def has_key(self, key: str) -> bool:
        """
        Check if key is registered (from IRegistry).
        
        Args:
            key: Registry key
            
        Returns:
            True if registered, False otherwise
        """
        ...

# =============================================================================
# STRATEGY INTERFACES (New Strategy Pattern)
# =============================================================================

@runtime_checkable
class ICachingStrategy(Protocol):
    """
    Generic caching strategy interface - works with ANY data type.
    
    Used by both modules and packages for caching strategies.
    """
    def get(self, key: str) -> Optional[Any]:
        """Get value from cache."""
        ...
    
    def set(self, key: str, value: Any) -> None:
        """Set value in cache."""
        ...
    
    def invalidate(self, key: str) -> None:
        """Invalidate cached value."""
        ...
    
    def clear(self) -> None:
        """Clear all cached values."""
        ...

@runtime_checkable
class IModuleHelperStrategy(Protocol):
    """
    Module helper strategy interface.
    
    Operations on a single module (loading, unloading, checking).
    """
    def load(self, module_path: str, package_helper: Any) -> ModuleType:
        """Load the module."""
        ...
    
    def unload(self, module_path: str) -> None:
        """Unload the module."""
        ...
    
    def check_importability(self, path: str) -> bool:
        """Check if module is importable."""
        ...

@runtime_checkable
class IPackageHelperStrategy(Protocol):
    """
    Package helper strategy interface.
    
    Operations on a single package (installing, uninstalling, checking).
    """
    def install(self, package_name: str) -> bool:
        """Install the package."""
        ...
    
    def uninstall(self, package_name: str) -> None:
        """Uninstall the package."""
        ...
    
    def check_installed(self, name: str) -> bool:
        """Check if package is installed."""
        ...
    
    def get_version(self, name: str) -> Optional[str]:
        """Get installed version."""
        ...

@runtime_checkable
class IModuleManagerStrategy(Protocol):
    """
    Module manager strategy interface.
    
    Orchestrates multiple modules (loading, hooks, error handling).
    """
    def load_module(self, module_path: str) -> ModuleType:
        """Load a module."""
        ...
    
    def unload_module(self, module_path: str) -> None:
        """Unload a module."""
        ...
    
    def install_hook(self) -> None:
        """Install import hook."""
        ...
    
    def uninstall_hook(self) -> None:
        """Uninstall import hook."""
        ...
    
    def handle_import_error(self, module_name: str) -> Optional[ModuleType]:
        """Handle import error."""
        ...

@runtime_checkable
class IPackageManagerStrategy(Protocol):
    """
    Package manager strategy interface.
    
    Orchestrates multiple packages (installation, discovery, policy).
    """
    def install_package(self, package_name: str, module_name: Optional[str] = None) -> bool:
        """Install a package."""
        ...
    
    def uninstall_package(self, package_name: str) -> None:
        """Uninstall a package."""
        ...
    
    def discover_dependencies(self) -> Dict[str, str]:
        """Discover dependencies."""
        ...
    
    def check_security_policy(self, package_name: str) -> Tuple[bool, str]:
        """Check security policy."""
        ...

# =============================================================================
# NEW PACKAGE STRATEGY TYPES (Redesigned Architecture)
# =============================================================================

@runtime_checkable
class IInstallExecutionStrategy(Protocol):
    """
    Installation execution strategy - HOW to execute installation.
    
    Defines the mechanism for actually installing packages (pip, wheel, cached, async).
    """
    def execute_install(self, package_name: str, policy_args: List[str]) -> Any:
        """Execute installation of a package."""
        ...
    
    def execute_uninstall(self, package_name: str) -> bool:
        """Execute uninstallation of a package."""
        ...

@runtime_checkable
class IInstallTimingStrategy(Protocol):
    """
    Installation timing strategy - WHEN to install packages.
    
    Defines when packages should be installed (on-demand, upfront, temporary, etc.).
    """
    def should_install_now(self, package_name: str, context: Any) -> bool:
        """Determine if package should be installed now."""
        ...
    
    def should_uninstall_after(self, package_name: str, context: Any) -> bool:
        """Determine if package should be uninstalled after use."""
        ...
    
    def get_install_priority(self, packages: List[str]) -> List[str]:
        """Get priority order for installing packages."""
        ...

@runtime_checkable
class IDiscoveryStrategy(Protocol):
    """
    Discovery strategy - HOW to discover dependencies.
    
    Defines how to find dependencies (from files, manifest, auto-detect).
    """
    def discover(self, project_root: Any) -> Dict[str, str]:
        """Discover dependencies from sources."""
        ...
    
    def get_source(self, import_name: str) -> Optional[str]:
        """Get the source of a discovered dependency."""
        ...

@runtime_checkable
class IPolicyStrategy(Protocol):
    """
    Policy strategy - WHAT can be installed (security/policy).
    
    Defines security policies and what packages are allowed/denied.
    """
    def is_allowed(self, package_name: str) -> Tuple[bool, str]:
        """Check if package is allowed to be installed."""
        ...
    
    def get_pip_args(self, package_name: str) -> List[str]:
        """Get pip arguments based on policy."""
        ...

@runtime_checkable
class IMappingStrategy(Protocol):
    """
    Mapping strategy - HOW to map import names to package names.
    
    Defines how to map import names (e.g., 'cv2') to package names (e.g., 'opencv-python').
    """
    def map_import_to_package(self, import_name: str) -> Optional[str]:
        """Map import name to package name."""
        ...
    
    def map_package_to_imports(self, package_name: str) -> List[str]:
        """Map package name to possible import names."""
        ...

# =============================================================================
# EXPORT ALL
# =============================================================================

__all__ = [
    # Interfaces/Protocols (Unified)
    'IPackageHelper',  # Merges: IPackageDiscovery, IPackageInstaller, IPackageCache, IConfigManager, IManifestLoader, IDependencyMapper
    'IModuleHelper',  # Merges: IModuleInstaller, IImportHook, IMetaPathFinder, IImportInterceptor, ILazyLoader, ILazyImporter, IWatchedRegistry, IBytecodeCache
    'IRuntime',  # Merges: IStateManager, IAdaptiveLearner, IIntelligentSelector, IMetricsCollector, IPerformanceMonitor, IMultiTierCache, IRegistry
    # Strategy Interfaces (New)
    'ICachingStrategy',
    'IModuleHelperStrategy',
    'IModuleManagerStrategy',
    'IPackageHelperStrategy',
    'IPackageManagerStrategy',
    # New Package Strategy Types
    'IInstallExecutionStrategy',
    'IInstallTimingStrategy',
    'IDiscoveryStrategy',
    'IPolicyStrategy',
    'IMappingStrategy',
]




"""
Module Operations Module

This module provides concrete implementations for module operations.
Main facade: XWModuleHelper extends AModuleHelper
"""

# Lazy imports to avoid circular dependencies
def __getattr__(name: str):
    if name == 'XWModuleHelper':
        from .facade import XWModuleHelper
        return XWModuleHelper
    if name == 'XWModule':  # Backward compatibility alias
        from .facade import XWModuleHelper
        return XWModuleHelper
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

__all__ = ['XWModuleHelper', 'XWModule']  # XWModule is backward compatibility alias



"""
Advanced Module Manager Strategy - Full features.

Company: eXonware.com
Author: Eng. Muhammad AlShehri
Email: connect@exonware.com

Generation Date: 15-Nov-2025

Advanced manager - with hooks, error handling, and full features.
"""

import sys
from types import ModuleType
from typing import Optional
from ...module.base import AModuleManagerStrategy
from ...contracts import ICachingStrategy, IModuleHelperStrategy
from ...package.base import APackageHelper
from ...module import importer_engine

class AdvancedManager(AModuleManagerStrategy):
    """
    Advanced manager - full features with hooks and error handling.
    
    Supports import hooks, error recovery, and all advanced features.
    """
    
    def __init__(
        self,
        package_name: str,
        package_helper: APackageHelper,
        caching: ICachingStrategy,
        helper: IModuleHelperStrategy
    ):
        """
        Initialize advanced manager.
        
        Args:
            package_name: Package name for isolation
            package_helper: Package helper instance
            caching: Caching strategy
            helper: Helper strategy
        """
        self._package_name = package_name
        self._package_helper = package_helper
        self._caching = caching
        self._helper = helper
        self._import_hook: Optional[importer_engine.LazyImportHook] = None
    
    def load_module(self, module_path: str) -> ModuleType:
        """
        Load a module.
        
        Args:
            module_path: Module path to load
            
        Returns:
            Loaded module
        """
        # Check cache first
        cached = self._caching.get(module_path)
        if cached is not None:
            return cached
        
        # Load using helper
        module = self._helper.load(module_path, self._package_helper)
        
        # Cache it
        self._caching.set(module_path, module)
        
        return module
    
    def unload_module(self, module_path: str) -> None:
        """
        Unload a module.
        
        Args:
            module_path: Module path to unload
        """
        self._helper.unload(module_path)
        self._caching.invalidate(module_path)
        
        # Also remove from sys.modules if present
        if module_path in sys.modules:
            del sys.modules[module_path]
    
    def install_hook(self) -> None:
        """Install import hook into sys.meta_path."""
        importer_engine.install_import_hook(self._package_name)
    
    def uninstall_hook(self) -> None:
        """Uninstall import hook from sys.meta_path."""
        importer_engine.uninstall_import_hook(self._package_name)
    
    def handle_import_error(self, module_name: str) -> Optional[ModuleType]:
        """
        Handle ImportError by attempting to install and re-import.
        
        Args:
            module_name: Name of module that failed to import
            
        Returns:
            Imported module if successful, None otherwise
        """
        if self._import_hook is None:
            self._import_hook = importer_engine.LazyImportHook(
                self._package_name,
                self._package_helper
            )
        return self._import_hook.handle_import_error(module_name)




"""
Dependency Mapper for package discovery.

This module contains DependencyMapper class extracted from lazy_core.py Section 1.
"""

import threading
from typing import Dict, List, Optional, Set, Tuple

# Lazy import to avoid circular dependency
def _get_logger():
    """Get logger (lazy import to avoid circular dependency)."""
    from ...common.logger import get_logger
    return get_logger("xwlazy.discovery")

logger = None  # Will be initialized on first use

# Import from spec_cache (same directory)
from .spec_cache import (
    _cached_stdlib_check,
    _spec_cache_get,
    _cache_spec_if_missing,
    get_stdlib_module_set,
)

_STDLIB_MODULE_SET = get_stdlib_module_set()

# Import from manifest (package/services - avoid circular import)
def get_manifest_loader():
    """Get manifest loader (lazy import to avoid circular dependency)."""
    from ...package.services.manifest import get_manifest_loader as _get_manifest_loader
    return _get_manifest_loader()

# Import from discovery (package/services - avoid circular import)
def get_lazy_discovery():
    """Get discovery instance."""
    from ...package.services.discovery import get_lazy_discovery as _get_lazy_discovery
    return _get_lazy_discovery()

class DependencyMapper:
    """
    Maps import names to package names using dynamic discovery.
    Optimized with caching to avoid repeated file I/O.
    """
    
    __slots__ = (
        '_discovery',
        '_package_import_mapping',
        '_import_package_mapping',
        '_cached',
        '_lock',
        '_package_name',
        '_manifest_generation',
        '_manifest_dependencies',
        '_manifest_signature',
        '_manifest_empty',
    )
    
    def __init__(self, package_name: str = 'default'):
        """Initialize dependency mapper."""
        self._discovery = None  # Lazy init to avoid circular imports
        self._package_import_mapping: Dict[str, List[str]] = {}
        self._import_package_mapping: Dict[str, str] = {}
        self._cached = False
        self._lock = threading.RLock()
        self._package_name = package_name
        self._manifest_generation = -1
        self._manifest_dependencies: Dict[str, str] = {}
        self._manifest_signature: Optional[Tuple[str, float, float]] = None
        self._manifest_empty = False

    def set_package_name(self, package_name: str) -> None:
        """Update the owning package name (affects manifest lookups)."""
        normalized = (package_name or 'default').strip().lower() or 'default'
        if normalized != self._package_name:
            self._package_name = normalized
            self._manifest_generation = -1
            self._manifest_dependencies = {}
    
    def _get_discovery(self):
        """Get discovery instance (lazy init)."""
        if self._discovery is None:
            self._discovery = get_lazy_discovery()
        return self._discovery
    
    def _ensure_mappings_cached(self) -> None:
        """Ensure mappings are cached (lazy initialization)."""
        if self._cached:
            return
        
        with self._lock:
            if self._cached:
                return
            
            discovery = self._get_discovery()
            self._package_import_mapping = discovery.get_package_import_mapping()
            self._import_package_mapping = discovery.get_import_package_mapping()
            self._cached = True
    
    def _ensure_manifest_cached(self, loader=None) -> None:
        if loader is None:
            loader = get_manifest_loader()
        signature = loader.get_manifest_signature(self._package_name)
        if signature == self._manifest_signature and (self._manifest_dependencies or self._manifest_empty):
            return
        
        shared = loader.get_shared_dependencies(self._package_name, signature)
        if shared is not None:
            self._manifest_generation = loader.generation
            self._manifest_signature = signature
            self._manifest_dependencies = shared
            self._manifest_empty = len(shared) == 0
            return

        manifest = loader.get_manifest(self._package_name)
        current_generation = loader.generation

        dependencies: Dict[str, str] = {}
        manifest_empty = True
        if manifest and manifest.dependencies:
            dependencies = {
                key.lower(): value
                for key, value in manifest.dependencies.items()
                if key and value
            }
            manifest_empty = False

        self._manifest_generation = current_generation
        self._manifest_signature = signature
        self._manifest_dependencies = dependencies
        self._manifest_empty = manifest_empty
    
    @staticmethod
    def _is_stdlib_or_builtin(module_name: str) -> bool:
        """Return True if the module is built-in or part of the stdlib."""
        root = module_name.split('.', 1)[0]
        needs_cache = False
        if module_name in _STDLIB_MODULE_SET or root in _STDLIB_MODULE_SET:
            return True
        if _cached_stdlib_check(module_name):
            needs_cache = True
        if needs_cache:
            _cache_spec_if_missing(module_name)
        return needs_cache

    DENY_LIST: Set[str] = {
        # POSIX-only modules that don't exist on Windows but try to auto-install
        "pwd",
        "grp",
        "spwd",
        "nis",
        "termios",
        "tty",
        "pty",
        "fcntl",
        # Windows-only internals
        "winreg",
        "winsound",
        "_winapi",
        "_dbm",
        # Internal optional modules that must never trigger auto-install
        "compression",
        "socks",
        "wimlib",
        # Optional dependencies with Python 2 compatibility shims (Python 3.8+ only)
        "inspect2",  # Python 2 compatibility shim, not needed on Python 3.8+
        "rich",      # Optional CLI enhancement for httpx, not required for core functionality
    }

    def _should_skip_auto_install(self, import_name: str) -> bool:
        """Determine whether an import should bypass lazy installation."""
        global logger
        if logger is None:
            logger = _get_logger()
        
        if self._is_stdlib_or_builtin(import_name):
            logger.debug("Skipping lazy install for stdlib module '%s'", import_name)
            return True

        if import_name in self.DENY_LIST:
            logger.debug("Skipping lazy install for denied module '%s'", import_name)
            return True

        return False

    def get_package_name(self, import_name: str) -> Optional[str]:
        """
        Get package name from import name.
        
        Priority order (manifest takes precedence):
        1. Skip checks (stdlib, deny list)
        2. Manifest dependencies (explicit user configuration - highest priority)
        3. Spec cache (module already exists - skip auto-install)
        4. Discovery mappings (automatic discovery)
        """
        if self._should_skip_auto_install(import_name):
            return None
        
        # Check manifest FIRST - explicit user configuration takes precedence
        loader = get_manifest_loader()
        generation_changed = self._manifest_generation != loader.generation
        manifest_uninitialized = not self._manifest_dependencies and not self._manifest_empty
        if generation_changed or manifest_uninitialized:
            self._ensure_manifest_cached(loader)
        manifest_hit = self._manifest_dependencies.get(import_name.lower())
        if manifest_hit:
            return manifest_hit
        
        # Check spec cache - if module already exists, skip auto-install
        if _spec_cache_get(import_name):
            return None

        self._ensure_mappings_cached()
        return self._import_package_mapping.get(import_name, import_name)
    
    def get_import_names(self, package_name: str) -> List[str]:
        """Get all possible import names for a package."""
        self._ensure_mappings_cached()
        return self._package_import_mapping.get(package_name, [package_name])
    
    def get_package_import_mapping(self) -> Dict[str, List[str]]:
        """Get complete package to import names mapping."""
        self._ensure_mappings_cached()
        return self._package_import_mapping.copy()
    
    def get_import_package_mapping(self) -> Dict[str, str]:
        """Get complete import to package name mapping."""
        self._ensure_mappings_cached()
        return self._import_package_mapping.copy()

__all__ = ['DependencyMapper']




"""
Configuration management for lazy loading system.

This module contains LazyInstallConfig which manages per-package lazy installation
configuration. Extracted from lazy_core.py Section 5.
"""

from typing import Dict, Optional
from ...common.services import LazyStateManager
from ...defs import LazyLoadMode, LazyInstallMode, LazyModeConfig
from ...defs import get_preset_mode

# Lazy import to avoid circular dependency
def _get_logger():
    """Get logger (lazy import to avoid circular dependency)."""
    from ...common.logger import get_logger
    return get_logger("xwlazy.config")

def _get_log_event():
    """Get log_event function (lazy import to avoid circular dependency)."""
    from ...common.logger import log_event
    return log_event

logger = None  # Will be initialized on first use
_log = None  # Will be initialized on first use

# Mode enum mapping - extracted from lazy_core.py
_MODE_ENUM_MAP = {
    # Core v1.0 modes
    "none": LazyInstallMode.NONE,
    "smart": LazyInstallMode.SMART,
    "full": LazyInstallMode.FULL,
    "clean": LazyInstallMode.CLEAN,
    "temporary": LazyInstallMode.TEMPORARY,
    "size_aware": LazyInstallMode.SIZE_AWARE,
    # Special purpose modes
    "interactive": LazyInstallMode.INTERACTIVE,
    "warn": LazyInstallMode.WARN,
    "disabled": LazyInstallMode.DISABLED,
    "dry_run": LazyInstallMode.DRY_RUN,
    # Legacy aliases
    "auto": LazyInstallMode.SMART,
    "on_demand": LazyInstallMode.SMART,
    "on-demand": LazyInstallMode.SMART,
    "lazy": LazyInstallMode.SMART,
}

class LazyInstallConfig:
    """Global configuration for lazy installation per package."""
    _configs: Dict[str, bool] = {}
    _modes: Dict[str, str] = {}
    _load_modes: Dict[str, LazyLoadMode] = {}
    _install_modes: Dict[str, LazyInstallMode] = {}
    _mode_configs: Dict[str, LazyModeConfig] = {}
    _initialized: Dict[str, bool] = {}
    _manual_overrides: Dict[str, bool] = {}
    
    @classmethod
    def set(
        cls,
        package_name: str,
        enabled: bool,
        mode: str = "auto",
        install_hook: bool = True,
        manual: bool = False,
        load_mode: Optional[LazyLoadMode] = None,
        install_mode: Optional[LazyInstallMode] = None,
        mode_config: Optional[LazyModeConfig] = None,
    ) -> None:
        """Enable or disable lazy installation for a specific package."""
        package_key = package_name.lower()
        state_manager = LazyStateManager(package_name)
        
        if manual:
            cls._manual_overrides[package_key] = True
            state_manager.set_manual_state(enabled)
        elif cls._manual_overrides.get(package_key):
            global logger
            if logger is None:
                logger = _get_logger()
            logger.debug(
                f"Lazy install config for {package_key} already overridden manually; skipping auto configuration."
            )
            return
        else:
            state_manager.set_manual_state(None)
        
        cls._configs[package_key] = enabled
        cls._modes[package_key] = mode
        
        # Handle two-dimensional mode configuration
        if mode_config:
            cls._mode_configs[package_key] = mode_config
            cls._load_modes[package_key] = mode_config.load_mode
            cls._install_modes[package_key] = mode_config.install_mode
        elif load_mode is not None or install_mode is not None:
            # Explicit mode specification
            if load_mode is None:
                load_mode = LazyLoadMode.AUTO  # Default
            if install_mode is None:
                install_mode = _MODE_ENUM_MAP.get(mode.lower(), LazyInstallMode.SMART)
            cls._load_modes[package_key] = load_mode
            cls._install_modes[package_key] = install_mode
            cls._mode_configs[package_key] = LazyModeConfig(
                load_mode=load_mode,
                install_mode=install_mode
            )
        else:
            # Legacy mode string - try to resolve to preset or default
            preset = get_preset_mode(mode)
            if preset:
                cls._mode_configs[package_key] = preset
                cls._load_modes[package_key] = preset.load_mode
                cls._install_modes[package_key] = preset.install_mode
            else:
                # Fallback to legacy behavior
                install_mode_enum = _MODE_ENUM_MAP.get(mode.lower(), LazyInstallMode.SMART)
                cls._load_modes[package_key] = LazyLoadMode.AUTO
                cls._install_modes[package_key] = install_mode_enum
                cls._mode_configs[package_key] = LazyModeConfig(
                    load_mode=LazyLoadMode.AUTO,
                    install_mode=install_mode_enum
                )
        
        cls._initialize_package(package_key, enabled, mode, install_hook=install_hook)
    
    @classmethod
    def _initialize_package(cls, package_key: str, enabled: bool, mode: str, install_hook: bool = True) -> None:
        """Initialize lazy installation for a specific package."""
        global logger, _log
        if logger is None:
            logger = _get_logger()
        if _log is None:
            _log = _get_log_event()
        
        # Deferred imports to avoid circular dependency
        from .install_registry import LazyInstallerRegistry
        from ...facade import (
            enable_lazy_install,
            disable_lazy_install,
            set_lazy_install_mode,
            enable_lazy_imports,
            install_import_hook,
            uninstall_import_hook,
            is_import_hook_installed,
            sync_manifest_configuration,
        )
        import asyncio
        
        if enabled:
            try:
                # Don't call enable_lazy_install() here - it would create infinite recursion
                # The config is already set by LazyInstallConfig.set() above
                
                # Use explicitly set install_mode from config, or derive from mode string
                # Check if install_mode was explicitly set by checking if package_key exists in _install_modes
                if package_key in cls._install_modes:
                    # install_mode was explicitly set in set() method, don't override it
                    mode_enum = cls._install_modes[package_key]
                else:
                    # Not explicitly set, derive from mode string
                    mode_enum = _MODE_ENUM_MAP.get(mode.lower(), LazyInstallMode.SMART)
                    set_lazy_install_mode(package_key, mode_enum)
                
                # Get load mode from config
                load_mode = cls.get_load_mode(package_key)
                
                # Enable lazy imports with appropriate load mode (skip if NONE mode)
                if load_mode != LazyLoadMode.NONE:
                    enable_lazy_imports(load_mode, package_name=package_key)
                
                # Enable async for modes that support it
                installer = LazyInstallerRegistry.get_instance(package_key)
                if installer and mode_enum in (LazyInstallMode.SMART, LazyInstallMode.FULL, LazyInstallMode.CLEAN, LazyInstallMode.TEMPORARY):
                    installer._async_enabled = True
                    installer._ensure_async_loop()
                    
                    # For FULL mode, install all dependencies on start
                    if mode_enum == LazyInstallMode.FULL:
                        loop = installer._async_loop
                        if loop:
                            asyncio.run_coroutine_threadsafe(installer.install_all_dependencies(), loop)
                
                if install_hook:
                    if not is_import_hook_installed(package_key):
                        install_import_hook(package_key)
                    _log("config", logger.info, f"✅ Lazy installation initialized for {package_key} (install_mode: {mode}, load_mode: {load_mode.value}, hook: installed)")
                else:
                    uninstall_import_hook(package_key)
                    _log("config", logger.info, f"✅ Lazy installation initialized for {package_key} (install_mode: {mode}, load_mode: {load_mode.value}, hook: disabled)")
                
                cls._initialized[package_key] = True
                sync_manifest_configuration(package_key)
            except ImportError as e:
                if logger is None:
                    logger = _get_logger()
                logger.warning(f"⚠️ Could not enable lazy install for {package_key}: {e}")
        else:
            try:
                disable_lazy_install(package_key)
            except ImportError:
                pass
            uninstall_import_hook(package_key)
            cls._initialized[package_key] = False
            _log("config", logger.info, f"❌ Lazy installation disabled for {package_key}")
            sync_manifest_configuration(package_key)
    
    @classmethod
    def is_enabled(cls, package_name: str) -> bool:
        """Check if lazy installation is enabled for a package."""
        return cls._configs.get(package_name.lower(), False)
    
    @classmethod
    def get_mode(cls, package_name: str) -> str:
        """Get the lazy installation mode for a package."""
        return cls._modes.get(package_name.lower(), "auto")
    
    @classmethod
    def get_mode_config(cls, package_name: str) -> Optional[LazyModeConfig]:
        """Get the full mode configuration for a package."""
        return cls._mode_configs.get(package_name.lower())
    
    @classmethod
    def get_load_mode(cls, package_name: str) -> LazyLoadMode:
        """Get the load mode for a package."""
        return cls._load_modes.get(package_name.lower(), LazyLoadMode.NONE)
    
    @classmethod
    def get_install_mode(cls, package_name: str) -> LazyInstallMode:
        """Get the install mode for a package."""
        return cls._install_modes.get(package_name.lower(), LazyInstallMode.NONE)
    
    @classmethod
    def set_install_mode(cls, package_name: str, mode: LazyInstallMode) -> None:
        """Set the install mode for a package."""
        package_key = package_name.lower()
        cls._install_modes[package_key] = mode
        # Update mode config if it exists
        if package_key in cls._mode_configs:
            mode_config = cls._mode_configs[package_key]
            cls._mode_configs[package_key] = LazyModeConfig(
                load_mode=mode_config.load_mode,
                install_mode=mode
            )




"""
Package Operations Facade

Main facade: XWPackageHelper extends APackageHelper
Provides concrete implementation for all package operations.
Uses strategy pattern for caching, helper, and manager strategies.
"""

import sys
import subprocess
import importlib
import importlib.util
from typing import Optional

from .base import APackageHelper
from .services import InstallerEngine, LazyInstaller
from .services.strategy_registry import StrategyRegistry
from ..common.cache import InstallationCache
from ..common.logger import get_logger

logger = get_logger("xwlazy.package.facade")

# Import strategy interfaces
from ..contracts import (
    ICachingStrategy,
    IPackageHelperStrategy,
    IPackageManagerStrategy,
    IInstallExecutionStrategy,
    IInstallTimingStrategy,
    IDiscoveryStrategy,
    IPolicyStrategy,
    IMappingStrategy,
)

# Import default strategies
from ..common.strategies import InstallationCacheWrapper
from .strategies import (
    # Execution strategies
    PipExecution,
    # Timing strategies
    SmartTiming,
    # Discovery strategies
    HybridDiscovery,
    # Policy strategies
    PermissivePolicy,
    # Mapping strategies
    ManifestFirstMapping,
)

class XWPackageHelper(APackageHelper):
    """
    Concrete implementation of APackageHelper.
    
    Provides simple, clean API for working with packages (what you pip install).
    Uses xwlazy's InstallationCache for persistent caching and LazyInstaller for installation.
    """
    
    def __init__(
        self,
        package_name: str = 'default',
        project_root: Optional[str] = None,
        *,
        # Legacy strategy injection (for backward compatibility)
        caching_strategy: Optional[ICachingStrategy] = None,
        helper_strategy: Optional[IPackageHelperStrategy] = None,
        manager_strategy: Optional[IPackageManagerStrategy] = None,
        # New strategy types
        execution_strategy: Optional[IInstallExecutionStrategy] = None,
        timing_strategy: Optional[IInstallTimingStrategy] = None,
        discovery_strategy: Optional[IDiscoveryStrategy] = None,
        policy_strategy: Optional[IPolicyStrategy] = None,
        mapping_strategy: Optional[IMappingStrategy] = None,
    ):
        """
        Initialize XW package helper.
        
        Args:
            package_name: Package name for isolation (defaults to 'default')
            project_root: Root directory of project (auto-detected if None)
            caching_strategy: Optional caching strategy. If None, uses InstallationCache.
            helper_strategy: Optional helper strategy (legacy, deprecated).
            manager_strategy: Optional manager strategy (legacy, deprecated).
            execution_strategy: Optional execution strategy. If None, uses PipExecution.
            timing_strategy: Optional timing strategy. If None, uses SmartTiming.
            discovery_strategy: Optional discovery strategy. If None, uses HybridDiscovery.
            policy_strategy: Optional policy strategy. If None, uses PermissivePolicy.
            mapping_strategy: Optional mapping strategy. If None, uses ManifestFirstMapping.
        """
        super().__init__(package_name, project_root)
        
        # Default strategies (legacy - deprecated, kept for backward compatibility)
        if caching_strategy is None:
            caching_strategy = InstallationCacheWrapper()
        # Legacy helper_strategy and manager_strategy are deprecated
        # They are kept for backward compatibility but not used
        
        # Check registry for stored strategies, otherwise use defaults
        if execution_strategy is None:
            execution_strategy = StrategyRegistry.get_package_strategy(package_name, 'execution')
            if execution_strategy is None:
                execution_strategy = PipExecution()
        if timing_strategy is None:
            timing_strategy = StrategyRegistry.get_package_strategy(package_name, 'timing')
            if timing_strategy is None:
                timing_strategy = SmartTiming()
        if discovery_strategy is None:
            discovery_strategy = StrategyRegistry.get_package_strategy(package_name, 'discovery')
            if discovery_strategy is None:
                discovery_strategy = HybridDiscovery(package_name, project_root)
        if policy_strategy is None:
            policy_strategy = StrategyRegistry.get_package_strategy(package_name, 'policy')
            if policy_strategy is None:
                policy_strategy = PermissivePolicy()
        if mapping_strategy is None:
            mapping_strategy = StrategyRegistry.get_package_strategy(package_name, 'mapping')
            if mapping_strategy is None:
                mapping_strategy = ManifestFirstMapping(package_name)
        
        # Store strategies
        self._caching = caching_strategy
        self._helper = helper_strategy  # Legacy, deprecated
        self._manager = manager_strategy  # Legacy, deprecated
        self._execution = execution_strategy
        self._timing = timing_strategy
        self._discovery = discovery_strategy
        self._policy = policy_strategy
        self._mapping = mapping_strategy
        
        # Legacy support (for backward compatibility)
        self._install_cache = InstallationCache()
        self._installer = None  # Lazy init to avoid circular imports
        self._install_engine = InstallerEngine(package_name)
    
    def _get_installer(self):
        """Get lazy installer instance (lazy init)."""
        if self._installer is None:
            self._installer = LazyInstaller(self._package_name)
        return self._installer
    
    def _check_importability(self, package_name: str) -> bool:
        """
        Check if package is importable.
        
        Uses importlib.util.find_spec to check if package can be imported.
        
        Args:
            package_name: Package name to check
            
        Returns:
            True if importable, False otherwise
        """
        try:
            spec = importlib.util.find_spec(package_name)
            return spec is not None
        except (ValueError, AttributeError, ImportError):
            return False
    
    def _check_persistent_cache(self, package_name: str) -> bool:
        """
        Check persistent cache for package installation status.
        
        Args:
            package_name: Package name to check
            
        Returns:
            True if found in persistent cache as installed, False otherwise
        """
        return self._install_cache.is_installed(package_name)
    
    def _mark_installed_in_persistent_cache(self, package_name: str) -> None:
        """
        Mark package as installed in persistent cache.
        
        Args:
            package_name: Package name to mark
        """
        version = self._get_installer()._get_installed_version(package_name)
        self._install_cache.mark_installed(package_name, version)
    
    def _mark_uninstalled_in_persistent_cache(self, package_name: str) -> None:
        """
        Mark package as uninstalled in persistent cache.
        
        Args:
            package_name: Package name to mark
        """
        self._install_cache.mark_uninstalled(package_name)
    
    def _run_install(self, *package_names: str) -> None:
        """
        Run pip install for packages.
        
        Uses execution strategy and timing strategy to determine when/how to install.
        
        Args:
            *package_names: Package names to install
            
        Raises:
            RuntimeError: If installation fails
        """
        if not package_names:
            return
        
        # Get policy args for each package
        policy_args_map = {}
        for package_name in package_names:
            # Check if should install now (timing strategy)
            if not self._timing.should_install_now(package_name, None):
                continue
            
            # Get pip args from policy strategy
            policy_args = self._policy.get_pip_args(package_name)
            policy_args_map[package_name] = policy_args
        
        # Execute installations using execution strategy
        for package_name, policy_args in policy_args_map.items():
            result = self._execution.execute_install(package_name, policy_args)
            
            # Handle result
            if hasattr(result, 'success') and result.success:
                with self._lock:
                    self._installed_packages.add(package_name)
                    self._uninstalled_packages.discard(package_name)
                    self._mark_installed_in_persistent_cache(package_name)
            else:
                with self._lock:
                    self._failed_packages.add(package_name)
                    error_msg = getattr(result, 'error', 'Unknown error') if hasattr(result, 'error') else str(result)
                    raise RuntimeError(f"Failed to install {package_name}: {error_msg}")
    
    def _run_uninstall(self, *package_names: str) -> None:
        """
        Run pip uninstall for packages.
        
        Uses execution strategy for uninstallation.
        
        Args:
            *package_names: Package names to uninstall
            
        Raises:
            RuntimeError: If uninstallation fails
        """
        if not package_names:
            return
        
        for package_name in package_names:
            # Check if should uninstall (timing strategy)
            if self._timing.should_uninstall_after(package_name, None):
                success = self._execution.execute_uninstall(package_name)
                if success:
                    with self._lock:
                        self._installed_packages.discard(package_name)
                        self._uninstalled_packages.add(package_name)
                        self._mark_uninstalled_in_persistent_cache(package_name)
                else:
                    raise RuntimeError(f"Failed to uninstall {package_name}")
    
    # Abstract methods from APackage that need implementation
    def _discover_from_sources(self) -> None:
        """Discover dependencies from all sources."""
        # Use discovery strategy
        deps = self._discovery.discover(self._project_root)
        # Convert to DependencyInfo format
        from ..defs import DependencyInfo
        for import_name, package_name in deps.items():
            self.discovered_dependencies[import_name] = DependencyInfo(
                import_name=import_name,
                package_name=package_name,
                source=self._discovery.get_source(import_name) or 'discovery',
                category='discovered'
            )
    
    def _is_cache_valid(self) -> bool:
        """Check if cached dependencies are still valid."""
        # Delegate to discovery strategy which has cache validation logic
        if hasattr(self._discovery, '_is_cache_valid'):
            return self._discovery._is_cache_valid()
        # Fallback: if discovery doesn't support cache validation, assume invalid
        return False
    
    def _add_common_mappings(self) -> None:
        """Add common import -> package mappings."""
        # Use mapping strategy to discover mappings
        # This is called during initialization to populate common mappings
        # The mapping strategy handles this internally
        pass
    
    def _update_file_mtimes(self) -> None:
        """Update file modification times for cache validation."""
        # Delegate to discovery strategy which tracks file modification times
        if hasattr(self._discovery, '_update_file_mtimes'):
            self._discovery._update_file_mtimes()
    
    # Strategy swapping methods
    def swap_cache_strategy(self, new_strategy: ICachingStrategy) -> None:
        """
        Swap cache strategy at runtime.
        
        Args:
            new_strategy: New caching strategy to use
        """
        self._caching = new_strategy
        # Update manager if it uses caching
        if hasattr(self._manager, '_caching'):
            self._manager._caching = new_strategy
    
    def swap_helper_strategy(self, new_strategy: IPackageHelperStrategy) -> None:
        """
        Swap helper/installer strategy at runtime.
        
        Args:
            new_strategy: New helper strategy to use
        """
        self._helper = new_strategy
        # Update manager if it uses helper
        if hasattr(self._manager, '_helper'):
            self._manager._helper = new_strategy
    
    def swap_manager_strategy(self, new_strategy: IPackageManagerStrategy) -> None:
        """
        Swap manager strategy at runtime.
        
        Args:
            new_strategy: New manager strategy to use
        """
        self._manager = new_strategy
    
    def swap_execution_strategy(self, new_strategy: IInstallExecutionStrategy) -> None:
        """
        Swap execution strategy at runtime.
        
        Args:
            new_strategy: New execution strategy to use
        """
        self._execution = new_strategy
    
    def swap_timing_strategy(self, new_strategy: IInstallTimingStrategy) -> None:
        """
        Swap timing strategy at runtime.
        
        Args:
            new_strategy: New timing strategy to use
        """
        self._timing = new_strategy
    
    def swap_discovery_strategy(self, new_strategy: IDiscoveryStrategy) -> None:
        """
        Swap discovery strategy at runtime.
        
        Args:
            new_strategy: New discovery strategy to use
        """
        self._discovery = new_strategy
    
    def swap_policy_strategy(self, new_strategy: IPolicyStrategy) -> None:
        """
        Swap policy strategy at runtime.
        
        Args:
            new_strategy: New policy strategy to use
        """
        self._policy = new_strategy
    
    def swap_mapping_strategy(self, new_strategy: IMappingStrategy) -> None:
        """
        Swap mapping strategy at runtime.
        
        Args:
            new_strategy: New mapping strategy to use
        """
        self._mapping = new_strategy
    
    def install_package(self, package_name: str, module_name: Optional[str] = None) -> bool:
        """
        Install a package.
        
        Uses timing strategy to determine if should install now,
        then uses execution strategy to perform installation.
        
        Args:
            package_name: Package name to install
            module_name: Optional module name (for mapping)
            
        Returns:
            True if installed successfully, False otherwise
        """
        # Map module name to package name if needed (using mapping strategy)
        if module_name and not package_name:
            package_name = self._mapping.map_import_to_package(module_name) or module_name
        
        # Check timing strategy
        if not self._timing.should_install_now(package_name, {'module_name': module_name}):
            return False
        
        # Check policy strategy
        allowed, reason = self._policy.is_allowed(package_name)
        if not allowed:
            raise RuntimeError(f"Package {package_name} blocked by policy: {reason}")
        
        # Get pip args from policy
        policy_args = self._policy.get_pip_args(package_name)
        
        # Execute installation
        result = self._execution.execute_install(package_name, policy_args)
        
        # Handle result
        if hasattr(result, 'success') and result.success:
            with self._lock:
                self._installed_packages.add(package_name)
                self._uninstalled_packages.discard(package_name)
                self._mark_installed_in_persistent_cache(package_name)
            return True
        else:
            with self._lock:
                self._failed_packages.add(package_name)
            return False
    
    def _check_security_policy(self, package_name: str):
        """Check security policy for package."""
        # Use policy strategy
        return self._policy.is_allowed(package_name)
    
    def _run_pip_install(self, package_name: str, args: list) -> bool:
        """Run pip install with arguments."""
        try:
            self._run_install(package_name)
            return True
        except (RuntimeError, subprocess.CalledProcessError, OSError) as e:
            logger.debug(f"Failed to install {package_name}: {e}")
            return False
        except Exception as e:
            # Catch-all for unexpected errors, but log them
            logger.warning(f"Unexpected error installing {package_name}: {e}")
            return False
    
    def is_cache_valid(self, key: str) -> bool:
        """Check if cache entry is still valid."""
        # Use caching strategy if it supports validation
        if hasattr(self._caching, 'is_valid') and self._caching is not None:
            return self._caching.is_valid(key)
        # Fallback: check if key exists in cache
        return self.get_cached(key) is not None
    
    # IConfigManager methods (delegate to LazyInstallConfig)
    def is_enabled(self, package_name: str) -> bool:
        """
        Check if lazy install is enabled for a package (from IConfigManager).
        
        This method delegates to LazyInstallConfig to avoid method name conflict
        with the instance method is_enabled().
        
        Args:
            package_name: Package name to check
            
        Returns:
            True if enabled, False otherwise
        """
        from .services.config_manager import LazyInstallConfig
        return LazyInstallConfig.is_enabled(package_name)
    
    def get_mode(self, package_name: str) -> str:
        """Get installation mode for a package (from IConfigManager)."""
        from .services.config_manager import LazyInstallConfig
        return LazyInstallConfig.get_mode(package_name)
    
    def get_load_mode(self, package_name: str):
        """Get load mode for a package (from IConfigManager)."""
        from .services.config_manager import LazyInstallConfig
        return LazyInstallConfig.get_load_mode(package_name)
    
    def get_install_mode(self, package_name: str):
        """Get install mode for a package (from IConfigManager)."""
        from .services.config_manager import LazyInstallConfig
        return LazyInstallConfig.get_install_mode(package_name)
    
    def get_mode_config(self, package_name: str):
        """Get full mode configuration for a package (from IConfigManager)."""
        from .services.config_manager import LazyInstallConfig
        return LazyInstallConfig.get_mode_config(package_name)




"""
Runtime Services Facade

Main facade: XWRuntimeHelper extends ARuntimeHelper
Provides concrete implementation for all runtime services.
"""

from typing import Dict, List, Optional, Tuple, Any
from .base import ARuntimeHelper
from ..defs import LazyLoadMode, LazyInstallMode

class XWRuntimeHelper(ARuntimeHelper):
    """
    Concrete implementation of ARuntimeHelper.
    
    Provides runtime services for state management, learning, monitoring, and caching.
    """
    
    def __init__(self):
        """Initialize XW runtime services."""
        super().__init__()
    
    # Abstract methods from ARuntime that need implementation
    def record_import(self, module_name: str, import_time: float) -> None:
        """Record an import event."""
        # TODO: Implement import recording
        pass
    
    def predict_next_imports(self, current_module: str, count: int = 3) -> List[str]:
        """Predict next likely imports based on patterns."""
        # TODO: Implement prediction logic
        return []
    
    def get_module_score(self, module_name: str) -> float:
        """Get priority score for a module."""
        # TODO: Implement scoring logic
        return 0.0
    
    def detect_load_level(
        self,
        module_count: int = 0,
        total_import_time: float = 0.0,
        import_count: int = 0,
        memory_usage_mb: float = 0.0
    ):
        """Detect current load level."""
        # TODO: Implement load level detection
        from .intelligent_selector import LoadLevel
        return LoadLevel.LIGHT
    
    def get_optimal_mode(self, load_level) -> Tuple[LazyLoadMode, LazyInstallMode]:
        """Get optimal mode for a load level."""
        # TODO: Implement mode selection
        return LazyLoadMode.AUTO, LazyInstallMode.SMART
    
    def update_mode_map(self, mode_map: Dict[Any, Tuple[LazyLoadMode, LazyInstallMode]]) -> None:
        """Update mode mapping with benchmark results."""
        # TODO: Implement mode map update
        pass
    
    def get_metric_stats(self, name: str) -> Dict[str, Any]:
        """Get statistics for a metric."""
        with self._lock:
            values = self._metrics.get(name, [])
            if not values:
                return {'count': 0, 'total': 0.0, 'average': 0.0, 'min': 0.0, 'max': 0.0}
            return {
                'count': len(values),
                'total': sum(values),
                'average': sum(values) / len(values),
                'min': min(values),
                'max': max(values)
            }
    
    def get_all_stats(self) -> Dict[str, Dict[str, Any]]:
        """Get statistics for all metrics."""
        with self._lock:
            return {name: self.get_metric_stats(name) for name in self._metrics.keys()}
    
    def get_performance_stats(self) -> Dict[str, Any]:
        """Get performance statistics."""
        with self._lock:
            return {
                'load_times': {k: {'count': len(v), 'total': sum(v), 'average': sum(v) / len(v) if v else 0.0} 
                              for k, v in self._load_times.items()},
                'access_counts': self._access_counts.copy(),
                'memory_usage': 0.0  # TODO: Implement memory usage tracking
            }
    
    def shutdown_multi_tier_cache(self) -> None:
        """Shutdown cache (flush L2, cleanup threads)."""
        # TODO: Implement cache shutdown
        pass




"""
Performance monitoring for lazy loading system.

This module contains LazyPerformanceMonitor extracted from lazy_core.py Section 4.
"""

from typing import Dict, Any

class LazyPerformanceMonitor:
    """Performance monitor for lazy loading operations."""
    
    __slots__ = ('_load_times', '_access_counts', '_memory_usage')
    
    def __init__(self):
        """Initialize performance monitor."""
        self._load_times: Dict[str, float] = {}
        self._access_counts: Dict[str, int] = {}
        self._memory_usage: Dict[str, Any] = {}
    
    def record_load_time(self, module: str, load_time: float) -> None:
        """Record module load time."""
        self._load_times[module] = load_time
    
    def record_access(self, module: str) -> None:
        """Record module access."""
        self._access_counts[module] = self._access_counts.get(module, 0) + 1
    
    def get_stats(self) -> Dict[str, Any]:
        """Get performance statistics."""
        return {
            'load_times': self._load_times.copy(),
            'access_counts': self._access_counts.copy(),
            'memory_usage': self._memory_usage.copy()
        }

__all__ = ['LazyPerformanceMonitor']




"""Lightweight helpers for registering host packages with xwlazy."""

from __future__ import annotations

import os
import sys
from typing import Iterable, Sequence, Tuple

# Lazy imports to avoid circular dependency
def _get_config_package_lazy_install_enabled():
    """Get config_package_lazy_install_enabled (lazy import to avoid circular dependency)."""
    from ...facade import config_package_lazy_install_enabled
    return config_package_lazy_install_enabled

def _get_install_import_hook():
    """Get install_import_hook (lazy import to avoid circular dependency)."""
    from ...facade import install_import_hook
    return install_import_hook

def _get_is_lazy_install_enabled():
    """Get is_lazy_install_enabled (lazy import to avoid circular dependency)."""
    from ...facade import is_lazy_install_enabled
    return is_lazy_install_enabled

def _get_register_lazy_module_methods():
    """Get register_lazy_module_methods (lazy import to avoid circular dependency)."""
    from ...facade import register_lazy_module_methods
    return register_lazy_module_methods

def _get_register_lazy_module_prefix():
    """Get register_lazy_module_prefix (lazy import to avoid circular dependency)."""
    from ...facade import register_lazy_module_prefix
    return register_lazy_module_prefix

# Note: LazyMetaPathFinder is implemented in module/meta_path_finder.py
# For now, create a placeholder
class LazyMetaPathFinder:
    """Placeholder for LazyMetaPathFinder - to be implemented in hooks domain."""
    def __init__(self, package_name: str):
        self.package_name = package_name
    
    def _enhance_classes_with_class_methods(self, module):
        """Placeholder method."""
        pass

_TRUTHY = {"1", "true", "yes", "on"}
_REGISTERED: dict[str, dict[str, Tuple[str, ...]]] = {}

def _normalized(items: Iterable[str]) -> Tuple[str, ...]:
    seen = []
    for item in items:
        if item not in seen:
            seen.append(item)
    return tuple(seen)

def register_host_package(
    package_name: str,
    module_prefixes: Iterable[str] = (),
    method_prefixes: Iterable[str] = (),
    method_names: Sequence[str] = ("encode", "decode"),
    auto_config: bool = True,
    env_var: str | None = None,
) -> None:
    """
    Register a host package (e.g., xwsystem) with xwlazy.

    Args:
        package_name: Host package name.
        module_prefixes: Prefixes that should be lazily wrapped.
        method_prefixes: Prefixes whose classes expose class-level helpers.
        method_names: Methods to expose at class level (default encode/decode).
        auto_config: If True, record lazy config but do not install hook yet.
        env_var: Optional environment variable to force enable (defaults to
                 ``{PACKAGE}_LAZY_INSTALL``).
    """
    package_name = package_name.lower()

    module_prefixes = _normalized(module_prefixes)
    method_prefixes = _normalized(method_prefixes)
    _REGISTERED[package_name] = {
        "module_prefixes": module_prefixes,
        "method_prefixes": method_prefixes,
    }

    for prefix in module_prefixes:
        _get_register_lazy_module_prefix()(prefix)

    for prefix in method_prefixes:
        _get_register_lazy_module_methods()(prefix, tuple(method_names))

    if auto_config:
        # Detect if lazy should be enabled (checks keyword, marker package, etc.)
        _get_config_package_lazy_install_enabled()(package_name, enabled=None, install_hook=False)
        
        # If detection found that lazy should be enabled, install the hook automatically
        if _get_is_lazy_install_enabled()(package_name):
            try:
                _get_install_import_hook()(package_name)
            except Exception:
                # Best-effort: package import must continue even if hook installation fails
                pass

    _apply_wrappers_for_loaded_modules(package_name, module_prefixes, method_prefixes)

    env_key = env_var or f"{package_name.upper()}_LAZY_INSTALL"
    flag = os.environ.get(env_key, "")
    if flag.strip().lower() in _TRUTHY:
        _get_config_package_lazy_install_enabled()(package_name, enabled=True)
        try:
            _get_install_import_hook()(package_name)
        except Exception:
            # Best-effort: package import must continue even if hook installation fails
            pass

def refresh_host_package(package_name: str) -> None:
    """Re-apply wrappers for a registered package."""
    data = _REGISTERED.get(package_name.lower())
    if not data:
        return
    _apply_wrappers_for_loaded_modules(
        package_name,
        data["module_prefixes"],
        data["method_prefixes"],
    )

def _apply_wrappers_for_loaded_modules(
    package_name: str,
    module_prefixes: Iterable[str],
    method_prefixes: Iterable[str],
) -> None:
    """Enhance already-imported modules so encode/decode helpers work immediately."""
    prefixes = _normalized((*module_prefixes, *method_prefixes))
    if not prefixes:
        return

    finder = LazyMetaPathFinder(package_name)
    for module_name, module in list(sys.modules.items()):
        if not isinstance(module_name, str) or module is None:
            continue
        if any(module_name.startswith(prefix) for prefix in prefixes):
            try:
                finder._enhance_classes_with_class_methods(module)  # type: ignore[attr-defined]
            except Exception:
                continue




"""
#exonware/xwlazy/src/exonware/xwlazy/defs.py

Company: eXonware.com
Author: Eng. Muhammad AlShehri
Email: connect@exonware.com
Generation Date: 10-Oct-2025

Type Definitions and Constants for Lazy Loading System

This module defines type definitions, constants, and TypedDict structures
for the lazy loading system following GUIDE_ARCH.md structure.
"""

from enum import Enum
from typing import TypedDict, Dict, List, Optional, Any, Tuple
from dataclasses import dataclass, field
from types import ModuleType

# =============================================================================
# ENUMS
# =============================================================================

class LazyLoadMode(Enum):
    """Controls lazy module loading behavior."""
    NONE = "none"           # Standard imports (no lazy loading)
    AUTO = "auto"           # Lazy loading enabled (deferred module loading)
    PRELOAD = "preload"     # Preload all modules on start
    BACKGROUND = "background"  # Load modules in background threads
    CACHED = "cached"       # Cache loaded modules but allow unloading
    # Superior performance modes
    TURBO = "turbo"         # Multi-tier cache + parallel preloading + bytecode caching
    ADAPTIVE = "adaptive"   # Self-optimizing with pattern learning
    HYPERPARALLEL = "hyperparallel"  # Maximum parallelism with multi-threading
    STREAMING = "streaming" # Asynchronous background loading with streaming
    ULTRA = "ultra"         # Aggressive optimizations (pre-compiled bytecode, mmap, etc.)
    INTELLIGENT = "intelligent"  # Automatically switches to fastest mode based on load level

class LazyInstallMode(Enum):
    """Lazy installation modes."""
    # Core modes
    NONE = "none"           # No auto-installation
    SMART = "smart"         # Install on first usage (on-demand) - replaces AUTO
    FULL = "full"           # Install all dependencies on start
    CLEAN = "clean"         # Install on usage + uninstall after completion
    TEMPORARY = "temporary"  # Always uninstall after use (more aggressive than CLEAN)
    SIZE_AWARE = "size_aware"  # Install small packages, skip large ones
    
    # Special purpose modes (kept for specific use cases)
    INTERACTIVE = "interactive"  # Ask user before installing
    WARN = "warn"           # Log warning but don't install (for monitoring)
    DISABLED = "disabled"   # Don't install anything (alias for NONE, more explicit)
    DRY_RUN = "dry_run"    # Show what would be installed but don't install

class PathType(Enum):
    """Path types for validation."""
    FILE = "file"
    DIRECTORY = "directory"
    UNKNOWN = "unknown"

class InstallStatus(Enum):
    """Installation status."""
    PENDING = "pending"
    RUNNING = "running"
    SUCCESS = "success"
    FAILED = "failed"
    SKIPPED = "skipped"

class LoadLevel(Enum):
    """Load level categories."""
    LIGHT = "light_load"
    MEDIUM = "medium_load"
    HEAVY = "heavy_load"
    ENTERPRISE = "enterprise_load"

# =============================================================================
# DATACLASSES
# =============================================================================

@dataclass
class DependencyInfo:
    """Information about a discovered dependency."""
    import_name: str
    package_name: str
    version: Optional[str] = None
    source: str = "unknown"
    category: str = "general"

@dataclass
class LazyModeConfig:
    """Two-dimensional lazy mode configuration combining load and install modes."""
    load_mode: LazyLoadMode = LazyLoadMode.NONE
    install_mode: LazyInstallMode = LazyInstallMode.NONE
    
    # Additional configuration options
    auto_uninstall_large: bool = False  # For AUTO_MODE behavior
    large_package_threshold_mb: float = 50.0  # Size threshold for SIZE_AWARE mode
    preload_priority: List[str] = field(default_factory=list)  # Priority modules for PRELOAD
    background_workers: int = 2  # Workers for BACKGROUND mode
    
    def __post_init__(self):
        """Normalize enum values."""
        if isinstance(self.load_mode, str):
            self.load_mode = LazyLoadMode(self.load_mode)
        if isinstance(self.install_mode, str):
            self.install_mode = LazyInstallMode(self.install_mode)

@dataclass
class InstallResult:
    """Result of an installation operation."""
    package_name: str
    success: bool
    status: InstallStatus
    error: Optional[str] = None
    version: Optional[str] = None
    source: Optional[str] = None  # "cache", "pip", "wheel", etc.

@dataclass
class LazyConfig:
    """Bridge configuration settings with the lazy package implementation."""
    packages: Tuple[str, ...] = field(
        default_factory=lambda: ("default",)
    )
    
    def __post_init__(self) -> None:
        self.packages = tuple(package.lower() for package in self.packages)

@dataclass(frozen=True)
class PackageManifest:
    """Resolved manifest data for a single package."""
    package: str
    dependencies: Dict[str, str] = field(default_factory=dict)
    watched_prefixes: Tuple[str, ...] = ()
    async_installs: bool = False
    async_workers: int = 1
    class_wrap_prefixes: Tuple[str, ...] = ()
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    def get_dependency(self, import_name: str) -> Optional[str]:
        """Return the declared package for the given import name."""
        if not import_name:
            return None
        direct = self.dependencies.get(import_name)
        if direct is not None:
            return direct
        # Case-insensitive fallback for convenience
        return self.dependencies.get(import_name.lower())

@dataclass(frozen=True)
class PackageData:
    """
    Immutable package data - same across all strategies.
    
    This data structure is used by all package caching, helper, and manager strategies.
    """
    name: str
    version: Optional[str] = None
    installed: bool = False
    install_time: Optional[float] = None
    access_count: int = 0
    install_mode: Optional['LazyInstallMode'] = None
    error: Optional[Exception] = None
    metadata: Dict[str, Any] = field(default_factory=dict)

@dataclass(frozen=True)
class ModuleData:
    """
    Immutable module data - same across all strategies.
    
    This data structure is used by all module caching, helper, and manager strategies.
    """
    path: str
    loaded_module: Optional['ModuleType'] = None
    loading: bool = False
    load_time: Optional[float] = None
    access_count: int = 0
    error: Optional[Exception] = None
    metadata: Dict[str, Any] = field(default_factory=dict)

# =============================================================================
# TYPE DEFINITIONS
# =============================================================================

class DependencyMapping(TypedDict, total=False):
    """Type definition for dependency mapping structure."""
    import_name: str
    package_name: str
    version: Optional[str]
    source: str
    category: str

class PackageStats(TypedDict, total=False):
    """Type definition for package statistics."""
    enabled: bool
    mode: str
    package_name: str
    installed_packages: List[str]
    failed_packages: List[str]
    total_installed: int
    total_failed: int

class LazyStatus(TypedDict, total=False):
    """Type definition for lazy mode status."""
    enabled: bool
    hook_installed: bool
    lazy_install_enabled: bool
    active: bool
    error: Optional[str]

# =============================================================================
# CONSTANTS
# =============================================================================

# Default configuration values
DEFAULT_LARGE_PACKAGE_THRESHOLD_MB: float = 50.0
DEFAULT_BACKGROUND_WORKERS: int = 2
DEFAULT_PRELOAD_PRIORITY: List[str] = []

# Common import -> package mappings (will be populated from discovery)
COMMON_IMPORT_MAPPINGS: Dict[str, str] = {
    # Common mappings that are frequently used
    'cv2': 'opencv-python',
    'PIL': 'Pillow',
    'yaml': 'PyYAML',
    'toml': 'toml',
    'pandas': 'pandas',
    'numpy': 'numpy',
    'scipy': 'scipy',
    'sklearn': 'scikit-learn',
    'matplotlib': 'matplotlib',
    'seaborn': 'seaborn',
    'requests': 'requests',
    'urllib3': 'urllib3',
    'bs4': 'beautifulsoup4',
    'lxml': 'lxml',
    'jinja2': 'Jinja2',
    'flask': 'Flask',
    'django': 'Django',
    'fastapi': 'fastapi',
    'pydantic': 'pydantic',
    'sqlalchemy': 'SQLAlchemy',
    'psycopg2': 'psycopg2-binary',
    'pymongo': 'pymongo',
    'redis': 'redis',
    'celery': 'celery',
    'boto3': 'boto3',
    'azure': 'azure-storage-blob',
    'google': 'google-cloud-storage',
    'tensorflow': 'tensorflow',
    'torch': 'torch',
    'transformers': 'transformers',
    'openai': 'openai',
    'anthropic': 'anthropic',
}

# Package discovery source names
DISCOVERY_SOURCE_PYPROJECT = "pyproject.toml"
DISCOVERY_SOURCE_REQUIREMENTS = "requirements.txt"
DISCOVERY_SOURCE_SETUP = "setup.py"
DISCOVERY_SOURCE_POETRY = "poetry.lock"
DISCOVERY_SOURCE_PIPFILE = "Pipfile"
DISCOVERY_SOURCE_MANIFEST = "manifest.json"

# Installation mode aliases
INSTALL_MODE_ALIASES: Dict[str, str] = {
    'auto': 'smart',
    'on_demand': 'smart',
    'on-demand': 'smart',
    'lazy': 'smart',
}

# Cache keys
CACHE_KEY_DEPENDENCIES = "dependencies"
CACHE_KEY_PACKAGE_INFO = "package_info"
CACHE_KEY_INSTALL_STATUS = "install_status"
CACHE_KEY_DISCOVERY_SOURCES = "discovery_sources"

# File patterns for discovery
PYPROJECT_PATTERN = "pyproject.toml"
REQUIREMENTS_PATTERN = "requirements*.txt"
SETUP_PATTERN = "setup.py"
POETRY_PATTERN = "poetry.lock"
PIPFILE_PATTERN = "Pipfile"

# =============================================================================
# PRESET MODE MAPPINGS
# =============================================================================

# Preset mode combinations for convenience
PRESET_MODES: Dict[str, LazyModeConfig] = {
    "none": LazyModeConfig(
        load_mode=LazyLoadMode.NONE,
        install_mode=LazyInstallMode.NONE
    ),
    "lite": LazyModeConfig(
        load_mode=LazyLoadMode.AUTO,
        install_mode=LazyInstallMode.NONE
    ),
    "smart": LazyModeConfig(
        load_mode=LazyLoadMode.AUTO,
        install_mode=LazyInstallMode.SMART
    ),
    "full": LazyModeConfig(
        load_mode=LazyLoadMode.AUTO,
        install_mode=LazyInstallMode.FULL
    ),
    "clean": LazyModeConfig(
        load_mode=LazyLoadMode.AUTO,
        install_mode=LazyInstallMode.CLEAN
    ),
    "temporary": LazyModeConfig(
        load_mode=LazyLoadMode.AUTO,
        install_mode=LazyInstallMode.TEMPORARY
    ),
    "size_aware": LazyModeConfig(
        load_mode=LazyLoadMode.AUTO,
        install_mode=LazyInstallMode.SIZE_AWARE
    ),
    "auto": LazyModeConfig(
        load_mode=LazyLoadMode.AUTO,
        install_mode=LazyInstallMode.SMART,
        auto_uninstall_large=True
    ),
}

def get_preset_mode(preset_name: str) -> Optional[LazyModeConfig]:
    """Get preset mode configuration by name."""
    return PRESET_MODES.get(preset_name.lower())

# =============================================================================
# EXPORT ALL
# =============================================================================

__all__ = [
    # Enums
    'LazyLoadMode',
    'LazyInstallMode',
    'PathType',
    'InstallStatus',
    'LoadLevel',
    # Dataclasses
    'DependencyInfo',
    'LazyModeConfig',
    'InstallResult',
    'LazyConfig',
    'PackageManifest',
    'PackageData',
    'ModuleData',
    # Type definitions
    'DependencyMapping',
    'PackageStats',
    'LazyStatus',
    # Constants
    'DEFAULT_LARGE_PACKAGE_THRESHOLD_MB',
    'DEFAULT_BACKGROUND_WORKERS',
    'DEFAULT_PRELOAD_PRIORITY',
    'COMMON_IMPORT_MAPPINGS',
    'DISCOVERY_SOURCE_PYPROJECT',
    'DISCOVERY_SOURCE_REQUIREMENTS',
    'DISCOVERY_SOURCE_SETUP',
    'DISCOVERY_SOURCE_POETRY',
    'DISCOVERY_SOURCE_PIPFILE',
    'DISCOVERY_SOURCE_MANIFEST',
    'INSTALL_MODE_ALIASES',
    'CACHE_KEY_DEPENDENCIES',
    'CACHE_KEY_PACKAGE_INFO',
    'CACHE_KEY_INSTALL_STATUS',
    'CACHE_KEY_DISCOVERY_SOURCES',
    'PYPROJECT_PATTERN',
    'REQUIREMENTS_PATTERN',
    'SETUP_PATTERN',
    'POETRY_PATTERN',
    'PIPFILE_PATTERN',
    # Preset modes
    'PRESET_MODES',
    'get_preset_mode',
]




"""
Package Strategies

All package strategy implementations.
"""

# Mapping strategies
from .package_mapping_manifest_first import ManifestFirstMapping
from .package_mapping_discovery_first import DiscoveryFirstMapping
from .package_mapping_hybrid import HybridMapping

# Policy strategies
from .package_policy_permissive import PermissivePolicy
from .package_policy_allow_list import AllowListPolicy
from .package_policy_deny_list import DenyListPolicy

# Timing strategies
from .package_timing_smart import SmartTiming
from .package_timing_full import FullTiming
from .package_timing_clean import CleanTiming
from .package_timing_temporary import TemporaryTiming

# Execution strategies
from .package_execution_pip import PipExecution
from .package_execution_wheel import WheelExecution
from .package_execution_cached import CachedExecution
from .package_execution_async import AsyncExecution

# Discovery strategies
from .package_discovery_file import FileBasedDiscovery
from .package_discovery_manifest import ManifestBasedDiscovery
from .package_discovery_hybrid import HybridDiscovery

__all__ = [
    # Mapping strategies
    'ManifestFirstMapping',
    'DiscoveryFirstMapping',
    'HybridMapping',
    # Policy strategies
    'PermissivePolicy',
    'AllowListPolicy',
    'DenyListPolicy',
    # Timing strategies
    'SmartTiming',
    'FullTiming',
    'CleanTiming',
    'TemporaryTiming',
    # Execution strategies
    'PipExecution',
    'WheelExecution',
    'CachedExecution',
    'AsyncExecution',
    # Discovery strategies
    'FileBasedDiscovery',
    'ManifestBasedDiscovery',
    'HybridDiscovery',
]



"""
Installation Utilities

Company: eXonware.com
Author: Eng. Muhammad AlShehri
Email: connect@exonware.com

Generation Date: 15-Nov-2025

Utility functions for installation operations.
"""

import os
import sys
import inspect
import subprocess
from pathlib import Path
from typing import Optional

def get_trigger_file() -> Optional[str]:
    """
    Get the file that triggered the import (from call stack).
    
    Returns:
        Filename that triggered the import, or None
    """
    try:
        # Walk up the call stack to find the first non-xwlazy file
        # Look for files in xwsystem, xwnode, xwdata, or user code
        for frame_info in inspect.stack():
            filename = frame_info.filename
            # Skip xwlazy internal files and importlib
            if ('xwlazy' not in filename and 
                'importlib' not in filename and 
                '<frozen' not in filename and
                filename.endswith('.py')):
                # Return just the filename, not full path
                basename = os.path.basename(filename)
                # If it's a serialization file, use that
                if 'serialization' in filename or 'formats' in filename:
                    # Extract the format name (e.g., bson.py -> BsonSerializer)
                    if basename.endswith('.py'):
                        basename = basename[:-3]  # Remove .py
                    return f"{basename.capitalize()}Serializer" if basename else None
                return basename
    except Exception:
        pass
    return None

def is_externally_managed() -> bool:
    """
    Check if Python environment is externally managed (PEP 668).
    
    Returns:
        True if environment is externally managed
    """
    marker_file = Path(sys.prefix) / "EXTERNALLY-MANAGED"
    return marker_file.exists()

def check_pip_audit_available() -> bool:
    """
    Check if pip-audit is available for vulnerability scanning.
    
    Returns:
        True if pip-audit is available
    """
    try:
        result = subprocess.run(
            [sys.executable, '-m', 'pip', 'list'],
            capture_output=True,
            text=True,
            timeout=5
        )
        return 'pip-audit' in result.stdout
    except Exception:
        return False

__all__ = ['get_trigger_file', 'is_externally_managed', 'check_pip_audit_available']




"""
Temporary Timing Strategy

Company: eXonware.com
Author: Eng. Muhammad AlShehri
Email: connect@exonware.com

Generation Date: 15-Nov-2025

Temporary timing - always uninstall after use (more aggressive than CLEAN).
"""

from typing import List, Any
from ...package.base import AInstallTimingStrategy

class TemporaryTiming(AInstallTimingStrategy):
    """
    Temporary timing strategy - always uninstalls after use (LazyInstallMode.TEMPORARY).
    
    More aggressive than CLEAN - always uninstalls packages after use.
    """
    
    def should_install_now(self, package_name: str, context: Any) -> bool:
        """
        Determine if package should be installed now.
        
        Temporary mode: Install when first needed.
        
        Args:
            package_name: Package name to check
            context: Context information (e.g., import error)
            
        Returns:
            True if should install now
        """
        return context is not None
    
    def should_uninstall_after(self, package_name: str, context: Any) -> bool:
        """
        Determine if package should be uninstalled after use.
        
        Temporary mode: Always uninstall after use.
        
        Args:
            package_name: Package name to check
            context: Context information (ignored)
            
        Returns:
            True (always uninstall)
        """
        return True
    
    def get_install_priority(self, packages: List[str]) -> List[str]:
        """
        Get priority order for installing packages.
        
        Temporary mode: Install in order requested.
        
        Args:
            packages: List of package names
            
        Returns:
            Priority-ordered list
        """
        return packages




"""
Package Services

Supporting services for package operations (not strategies).
"""

from .install_result import InstallStatus, InstallResult
from .install_policy import LazyInstallPolicy
from .install_utils import get_trigger_file, is_externally_managed, check_pip_audit_available
from .install_registry import LazyInstallerRegistry
from .installer_engine import InstallerEngine
from .async_install_handle import AsyncInstallHandle
from .lazy_installer import LazyInstaller
from .strategy_registry import StrategyRegistry

# Config and manifest services
from .config_manager import LazyInstallConfig
from .manifest import (
    PackageManifest,
    LazyManifestLoader,
    get_manifest_loader,
    refresh_manifest_cache,
    sync_manifest_configuration,
    _normalize_prefix,
)

# Discovery service
from .discovery import LazyDiscovery, get_lazy_discovery

# State management and keyword detection (moved to common/services)
from ...common.services import (
    LazyStateManager,
    enable_keyword_detection,
    is_keyword_detection_enabled,
    get_keyword_detection_keyword,
    check_package_keywords,
    _detect_lazy_installation,
    _detect_meta_info_mode,
)

# Host package registration
from .host_packages import (
    register_host_package,
    refresh_host_package,
)

__all__ = [
    # Install services
    'InstallStatus',
    'InstallResult',
    'LazyInstallPolicy',
    'LazyInstallerRegistry',
    'InstallerEngine',
    'AsyncInstallHandle',
    'LazyInstaller',
    'StrategyRegistry',
    'get_trigger_file',
    'is_externally_managed',
    'check_pip_audit_available',
    # Config and manifest
    'LazyInstallConfig',
    'PackageManifest',
    'LazyManifestLoader',
    'get_manifest_loader',
    'refresh_manifest_cache',
    'sync_manifest_configuration',
    '_normalize_prefix',
    # Discovery
    'LazyDiscovery',
    'get_lazy_discovery',
    # State management
    'LazyStateManager',
    # Keyword detection
    'enable_keyword_detection',
    'is_keyword_detection_enabled',
    'get_keyword_detection_keyword',
    'check_package_keywords',
    '_detect_lazy_installation',
    '_detect_meta_info_mode',
    # Host package registration
    'register_host_package',
    'refresh_host_package',
]




"""
Pip Execution Strategy

Company: eXonware.com
Author: Eng. Muhammad AlShehri
Email: connect@exonware.com

Generation Date: 15-Nov-2025

Pip execution strategy - direct pip install.
"""

import sys
import subprocess
from typing import List, Any
from ...package.base import AInstallExecutionStrategy
from ...package.services.install_result import InstallResult, InstallStatus

class PipExecution(AInstallExecutionStrategy):
    """
    Pip execution strategy - installs packages directly using pip.
    
    This is the default execution strategy that uses pip install.
    """
    
    def execute_install(self, package_name: str, policy_args: List[str]) -> Any:
        """
        Execute installation using pip.
        
        Args:
            package_name: Package name to install
            policy_args: Policy arguments (index URLs, trusted hosts, etc.)
            
        Returns:
            InstallResult with success status
        """
        try:
            pip_args = [sys.executable, '-m', 'pip', 'install']
            if policy_args:
                pip_args.extend(policy_args)
            pip_args.append(package_name)
            
            result = subprocess.run(
                pip_args,
                capture_output=True,
                text=True,
                check=True
            )
            
            if result.returncode == 0:
                return InstallResult(
                    package_name=package_name,
                    success=True,
                    status=InstallStatus.SUCCESS,
                    source="pip"
                )
            else:
                return InstallResult(
                    package_name=package_name,
                    success=False,
                    status=InstallStatus.FAILED,
                    error=result.stderr or "Unknown error"
                )
        except subprocess.CalledProcessError as e:
            return InstallResult(
                package_name=package_name,
                success=False,
                status=InstallStatus.FAILED,
                error=e.stderr or str(e)
            )
        except Exception as e:
            return InstallResult(
                package_name=package_name,
                success=False,
                status=InstallStatus.FAILED,
                error=str(e)
            )
    
    def execute_uninstall(self, package_name: str) -> bool:
        """
        Execute uninstallation using pip.
        
        Args:
            package_name: Package name to uninstall
            
        Returns:
            True if successful, False otherwise
        """
        try:
            result = subprocess.run(
                [sys.executable, '-m', 'pip', 'uninstall', '-y', package_name],
                capture_output=True,
                text=True,
                check=True
            )
            return result.returncode == 0
        except Exception:
            return False

